# Generated by Django 3.2.13 on 2022-05-05 22:24

import datetime
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DeviceFingerprint',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('browser', models.CharField(max_length=32)),
                ('browser_major_version', models.IntegerField()),
                ('device_category', models.CharField(max_length=32)),
                ('os', models.CharField(max_length=32)),
                ('platform', models.CharField(max_length=32)),
                ('screen_height', models.IntegerField()),
                ('screen_width', models.IntegerField()),
                ('user_agent', models.TextField()),
                ('viewport_height', models.IntegerField()),
                ('viewport_width', models.IntegerField()),
            ],
            options={
                'db_table': 'djspoofer_device_fingerprint',
            },
        ),
        migrations.CreateModel(
            name='Fingerprint',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
            options={
                'db_table': 'djspoofer_fingerprint',
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='Geolocation',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('city', models.CharField(max_length=64)),
                ('country', models.CharField(max_length=2)),
                ('isp', models.CharField(blank=True, max_length=64, null=True)),
            ],
            options={
                'db_table': 'djspoofer_geolocation',
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='H2Fingerprint',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('browser', models.CharField(max_length=32)),
                ('browser_min_major_version', models.IntegerField(default=0)),
                ('browser_max_major_version', models.IntegerField(default=999)),
                ('os', models.CharField(max_length=32)),
                ('header_table_size', models.IntegerField()),
                ('enable_push', models.BooleanField(blank=True, null=True)),
                ('max_concurrent_streams', models.IntegerField(blank=True, null=True)),
                ('initial_window_size', models.IntegerField()),
                ('max_frame_size', models.IntegerField(blank=True, null=True)),
                ('max_header_list_size', models.IntegerField(blank=True, null=True)),
                ('psuedo_header_order', models.TextField()),
                ('window_update_increment', models.IntegerField()),
                ('priority_stream_id', models.IntegerField()),
                ('priority_exclusive', models.BooleanField()),
                ('priority_depends_on_id', models.IntegerField()),
                ('priority_weight', models.IntegerField()),
            ],
            options={
                'db_table': 'djspoofer_h2_fingerprint',
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='IntoliFingerprint',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('browser', models.CharField(max_length=32)),
                ('browser_major_version', models.IntegerField()),
                ('device_category', models.CharField(max_length=32)),
                ('os', models.CharField(max_length=32)),
                ('platform', models.CharField(max_length=32)),
                ('screen_height', models.IntegerField()),
                ('screen_width', models.IntegerField()),
                ('user_agent', models.TextField()),
                ('viewport_height', models.IntegerField()),
                ('viewport_width', models.IntegerField()),
                ('weight', models.DecimalField(decimal_places=24, max_digits=25)),
            ],
            options={
                'db_table': 'djspoofer_intoli_fingerprint',
                'ordering': ['-weight'],
            },
        ),
        migrations.CreateModel(
            name='IP',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('city', models.CharField(max_length=64)),
                ('country', models.CharField(max_length=2)),
                ('isp', models.CharField(max_length=64)),
                ('address', models.GenericIPAddressField()),
            ],
            options={
                'db_table': 'djspoofer_ip',
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='Proxy',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('url', models.TextField(unique=True)),
                ('mode', models.IntegerField(choices=[(10, 'GENERAL'), (20, 'ROTATING'), (30, 'STICKY')], default=10)),
                ('country', models.CharField(blank=True, max_length=3, null=True)),
                ('city', models.CharField(blank=True, max_length=64, null=True)),
                ('last_used', models.DateTimeField(blank=True, null=True)),
                ('used_count', models.IntegerField(default=0)),
                ('cooldown', models.DurationField(default=datetime.timedelta(seconds=600))),
            ],
            options={
                'db_table': 'djspoofer_proxy',
                'ordering': ['url'],
            },
        ),
        migrations.CreateModel(
            name='TLSFingerprint',
            fields=[
                ('oid', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('browser', models.CharField(max_length=32)),
                ('browser_min_major_version', models.IntegerField(default=0)),
                ('browser_max_major_version', models.IntegerField(default=999)),
                ('os', models.CharField(max_length=32)),
                ('extensions', models.IntegerField()),
                ('ciphers', models.TextField()),
            ],
            options={
                'db_table': 'djspoofer_tls_fingerprint',
                'ordering': ['-created'],
            },
        ),
        migrations.AddIndex(
            model_name='tlsfingerprint',
            index=models.Index(fields=['browser'], name='tls_fp_browser'),
        ),
        migrations.AddIndex(
            model_name='ip',
            index=models.Index(fields=['city'], name='ip_fp_city'),
        ),
        migrations.AddIndex(
            model_name='ip',
            index=models.Index(fields=['country'], name='ip_fp_country'),
        ),
        migrations.AddIndex(
            model_name='ip',
            index=models.Index(fields=['isp'], name='ip_fp_isp'),
        ),
        migrations.AddIndex(
            model_name='intolifingerprint',
            index=models.Index(fields=['browser'], name='idx_intoli_fp_browser'),
        ),
        migrations.AddIndex(
            model_name='intolifingerprint',
            index=models.Index(fields=['device_category'], name='idx_intoli_fp_device_category'),
        ),
        migrations.AddIndex(
            model_name='intolifingerprint',
            index=models.Index(fields=['os'], name='idx_intoli_fp_os'),
        ),
        migrations.AddIndex(
            model_name='intolifingerprint',
            index=models.Index(fields=['platform'], name='idx_intoli_fp_platform'),
        ),
        migrations.AddField(
            model_name='fingerprint',
            name='_h2_fingerprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='fingerprints', to='djspoofer.h2fingerprint'),
        ),
        migrations.AddField(
            model_name='fingerprint',
            name='_tls_fingerprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='fingerprints', to='djspoofer.tlsfingerprint'),
        ),
        migrations.AddField(
            model_name='fingerprint',
            name='device_fingerprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='fingerprints', to='djspoofer.devicefingerprint'),
        ),
        migrations.AddField(
            model_name='fingerprint',
            name='geolocation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='fingerprints', to='djspoofer.geolocation'),
        ),
        migrations.AddField(
            model_name='fingerprint',
            name='ip',
            field=models.ManyToManyField(to='djspoofer.IP'),
        ),
        migrations.AddIndex(
            model_name='devicefingerprint',
            index=models.Index(fields=['browser'], name='idx_device_fp_browser'),
        ),
        migrations.AddIndex(
            model_name='devicefingerprint',
            index=models.Index(fields=['device_category'], name='idx_device_fp_device_category'),
        ),
        migrations.AddIndex(
            model_name='devicefingerprint',
            index=models.Index(fields=['platform'], name='idx_device_fp_platform'),
        ),
    ]



from djstarter import exceptions


class ProxyRackError(exceptions.AppError):
    """All Exceptions in the ProxyRack App inherit from this class"""


class ProxyRackApiError(exceptions.ApiError):
    """All Exceptions in the ProxyRack Api inherit from this class"""


class ProxyError(exceptions.ApiError):
    """All Proxy Exceptions"""


class ProxyNotAuthenticated(ProxyError):
    """
    Proxy Not Found - 407

    This error means that there is something wrong with your authentication credentials.

    If you are using IP whitelisting, make sure that you are connecting to your proxies from the same IP address that
    you have whitelisted in your user dashboard.

    If you are using username and password authentication, please make sure that you are using your API key as your
    password and not the password you use to log in to your Proxyrack account.


    """


class GeoLocationNotFound(ProxyError):
    """
    Geo Location Not Found - 560

    This means there are no available exit nodes for your GEO location request.

    Please double check you are using the correct country code.

    If you are using a correct country code and you are still receiving this error then you will need to try again
    later or remove the GEO targeting parameters.
    """


class ProxyUnreachable(ProxyError):
    """
    Proxy Unreachable - 561

    You will receive this error if the exit node was not reachable.

    In the odd circumstance that this happens, simply retry your connection.
    """


class ProxyNotFound(ProxyError):
    """
    Proxy Not Found - 562

    This means that we were unable to find an exit node for your request.

    If you are using any of our GEO location targeting features it may mean that there are no available nodes
    for your request.

    To avoid this simply remove your targeting features and try again.
    """


class ProxyNotOnline(ProxyError):
    """
    Proxy Not Found - 564

    This means that the proxy assigned to your session is no longer online and the request could not be completed.

    You will need to release the sticky session and retry your connection.

    Alternatively, you could wait to see if the exit node comes back online and retry again.
    """



import logging
from ssl import TLSVersion

import httpx
from django.conf import settings
from djstarter.clients import RetryClient

from djspoofer import utils
from djspoofer.models import Fingerprint
from djspoofer.remote.proxyrack import backends

logger = logging.getLogger(__name__)


class DesktopClient(RetryClient, backends.ProxyRackProxyBackend):
    def __init__(self, fingerprint=None, proxy_enabled=True, *args, **kwargs):
        self._proxy_enabled = proxy_enabled
        self.fingerprint = fingerprint or Fingerprint.objects.random_desktop()
        logger.info(f'Starting session with fingerprint: {self.fingerprint}')
        self.user_agent = self.fingerprint.device_fingerprint.user_agent
        super().__init__(
            headers=self.init_headers(),
            http2=True,
            proxies=self._get_proxies() if self._proxy_enabled else dict(),
            verify=self._get_ssl_context(),
            *args,
            **kwargs
        )

    def init_headers(self):
        return {
            # 'accept': 'application/json',
            'h2-fingerprint-id': self.fingerprint.h2_fingerprint.oid_str
        }

    def _get_proxies(self):
        if settings.PROXY_URL == 'False':
            return dict()
        proxy_url = settings.PROXY_URL or self.get_proxy_url(self.fingerprint)
        return utils.proxy_dict(proxy_url)

    def _get_ssl_context(self):
        tls_fingerprint = self.fingerprint.tls_fingerprint

        context = httpx.create_ssl_context(http2=True, verify=settings.SSL_VERIFY)
        context.minimum_version = TLSVersion.TLSv1_2
        context.set_ciphers(tls_fingerprint.ciphers)
        context.options = tls_fingerprint.extensions

        return context

    def send(self, *args, **kwargs):
        self.headers.pop('Accept-Encoding', None)
        self.headers.pop('Connection', None)
        return super().send(*args, **kwargs)


class DesktopChromeClient(DesktopClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ua_parser = utils.UserAgentParser(self.user_agent)

    def init_headers(self):
        return super().init_headers() | {
            'user-agent': self.user_agent,
        }

    @property
    def sec_ch_ua(self):
        version = self.ua_parser.browser_major_version
        return f'" Not;A Brand";v="99", "Google Chrome";v="{version}", "Chromium";v="{version}"'

    @property
    def sec_ch_ua_mobile(self):
        return '?0'

    @property
    def sec_ch_ua_platform(self):
        platform = self.ua_parser.os
        return f'"{platform}"'


class DesktopFirefoxClient(DesktopClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init_headers(self):
        return super().init_headers() | {
            'User-Agent': self.user_agent,
        }




from django.core.management.base import BaseCommand
from httpx import Client

from djspoofer.clients import DesktopChromeClient


class Command(BaseCommand):
    help = 'Test Proxies'

    def add_arguments(self, parser):
        parser.add_argument(
            "--url",
            required=True,
            type=str,
            help="Target URL for proxies",
        )
        parser.add_argument(
            "--proxy-enabled",
            required=False,
            type=bool,
            default=True,
            help="Proxy Enabled",
        )

    def handle(self, *args, **kwargs):
        url = kwargs['url']
        try:
            with DesktopChromeClient(proxy_enabled=kwargs.get('proxy_enabled', True)) as client:
                client.get(url)

        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successful GET for "{url}"'))



from django.core.management.base import BaseCommand

from djspoofer import const, utils
from djspoofer.models import Proxy, Fingerprint, DeviceFingerprint, IntoliFingerprint
from djspoofer.remote.intoli import tasks


class Command(BaseCommand):
    help = 'DJ Spoofer Init'

    def add_arguments(self, parser):
        parser.add_argument(
            "--proxy-url",
            required=True,
            type=str,
            help="Rotating Proxy Url (Example: premium.residential.proxyrack.net:10000)",
        )
        parser.add_argument(
            "--fingerprint-count",
            required=True,
            type=int,
            help="Fingerprint Count (Default 50)",
        )

    def handle(self, *args, **kwargs):
        try:
            self.create_h2_fingerprints()
            self.create_rotating_proxy(kwargs['proxy_url'])
            self.store_intoli_fingerprints()
            self.create_fingerprints(count=kwargs['fingerprint_count'])
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully Initialized DJ Spoofer'))

    def create_h2_fingerprints(self):
        utils.h2_hash_to_h2_fingerprint(
            os='Windows',
            browser='Chrome',
            h2_hash='1:65536;3:1000;4:6291456;6:262144|15663105|1:1:0:256|m,a,s,p',
            browser_min_major_version=70,
            browser_max_major_version=110,
        )
        utils.h2_hash_to_h2_fingerprint(
            os='Linux',
            browser='Chrome',
            h2_hash='1:65536;3:1000;4:6291456;6:262144|15663105|1:1:0:256|m,a,s,p',
            browser_min_major_version=70,
            browser_max_major_version=110,
        )
        utils.h2_hash_to_h2_fingerprint(
            os='Windows',
            browser='Edge',
            h2_hash='1:65536;3:1000;4:6291456;6:262144|15663105|1:1:0:256|m,a,s,p',
            browser_min_major_version=70,
            browser_max_major_version=110,
        )
        utils.h2_hash_to_h2_fingerprint(
            os='Windows',
            browser='Firefox',
            h2_hash='1:65536;4:131072;5:16384|12517377|15:0:13:42|m,p,a,s',
            browser_min_major_version=70,
            browser_max_major_version=110,
        )
        utils.h2_hash_to_h2_fingerprint(
            os='Linux',
            browser='Firefox',
            h2_hash='1:65536;4:131072;5:16384|12517377|15:0:13:42|m,p,a,s',
            browser_min_major_version=70,
            browser_max_major_version=110,
        )
        self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully Created H2 Fingerprints'))

    def create_rotating_proxy(self, proxy_url):
        proxy = Proxy.objects.create_rotating_proxy(url=proxy_url)
        self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully Created Rotating Proxy: {proxy}'))

    def store_intoli_fingerprints(self):
        tasks.get_profiles(max_profiles=150, desktop_only=True, os_list=const.SUPPORTED_OS)
        self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully Created Intoli Fingerprints'))

    def create_fingerprints(self, count):
        for i_fp in IntoliFingerprint.objects.weighted_n_desktop(count=count):
            ua_parser = utils.UserAgentParser(i_fp.user_agent)
            Fingerprint.objects.create(
                device_fingerprint=DeviceFingerprint.objects.create(
                    browser=ua_parser.browser,
                    browser_major_version=ua_parser.browser_major_version,
                    device_category=i_fp.device_category,
                    os=ua_parser.os,
                    platform=i_fp.platform,
                    screen_height=i_fp.screen_height,
                    screen_width=i_fp.screen_width,
                    user_agent=i_fp.user_agent,
                    viewport_height=i_fp.viewport_height,
                    viewport_width=i_fp.viewport_width,
                )
            )
        self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully Created {count} Fingerprints'))



import json
from importlib.resources import open_text
from unittest import mock

from django.test import TestCase
from httpx import Request

from djspoofer.models import IntoliFingerprint
from djspoofer.remote.intoli import tasks, intoli_api


class GetProfilesTaskTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.intoli.resources', 'user-agents.json') as user_agents_json:
            cls.r_data = json.loads(user_agents_json.read())

    @mock.patch.object(intoli_api, 'get_profiles')
    def test_ok(self, get_profiles):
        get_profiles.return_value = intoli_api.GetProfilesResponse(self.r_data)

        tasks.get_profiles()
        self.assertEquals(IntoliFingerprint.objects.count(), 5)

    @mock.patch.object(intoli_api, 'get_profiles')
    def test_desktop_only_and_os_list(self, get_profiles):
        get_profiles.return_value = intoli_api.GetProfilesResponse(self.r_data)

        tasks.get_profiles(desktop_only=True, os_list='Windows')
        self.assertEquals(IntoliFingerprint.objects.count(), 2)



USER_AGENT_BAD_CHARS = '\"\''
USER_AGENT_BLACKLIST = ('coc_coc_browser', 'QQBrowser', 'WeChat', 'YaBrowser', 'zh-CN', 'zh_CN', 'zhihu')
USER_AGENT_LEN_RANGE = (40, 160)
DEVICE_CATEGORY_LEN_RANGE = (4, 16)
PLATFORM_LEN_RANGE = (4, 16)

MAX_PROFILES = 150



import logging

from django.conf import settings
from djstarter import decorators

from .exceptions import IncolumitasError

logger = logging.getLogger(__name__)

BASE_URL = settings.INCOLUMITAS_TCPIP_API_BASE_URL


@decorators.wrap_exceptions(raise_as=IncolumitasError)
def tcpip_fingerprint(client):
    url = f'{BASE_URL}/classify'

    params = {
        'by_ip': '1',
    }

    r = client.get(url, params=params)
    r.raise_for_status()
    return TCPIPFingerprintResponse(r.json())


class TCPIPFingerprintResponse:
    class Guess:
        def __init__(self, data):
            self.os = data['os']
            self.score = data['score']

    def __init__(self, data):
        self.best_n_guesses = [self.Guess(g) for g in data['bestNGuesses']]
        self.ip = data['ip']
        self.vpn_detected = data['vpn_detected']

    @property
    def top_guess(self):
        return self.best_n_guesses[0]



from django.core.management.base import BaseCommand

from djspoofer.models import Proxy, Fingerprint, H2Fingerprint, IntoliFingerprint


class Command(BaseCommand):
    help = 'DJ Spoofer Reset'

    def handle(self, *args, **kwargs):
        try:
            self.delete_h2_fingerprints()
            self.delete_proxies()
            self.delete_intoli_fingerprints()
            self.delete_fingerprints()
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully Reset DJ Spoofer'))

    def delete_h2_fingerprints(self):
        deleted, _ = H2Fingerprint.objects.all().delete()
        self.stdout.write(self.style.MIGRATE_LABEL(f'Deleted H2 Fingerprints: {deleted}'))

    def delete_proxies(self):
        deleted, _ = Proxy.objects.all().delete()
        self.stdout.write(self.style.MIGRATE_LABEL(f'Deleted Proxies: {deleted}'))

    def delete_intoli_fingerprints(self):
        deleted, _ = IntoliFingerprint.objects.all().delete()
        self.stdout.write(self.style.MIGRATE_LABEL(f'Deleted Intoli Fingerprints: {deleted}'))

    def delete_fingerprints(self):
        deleted, _ = Fingerprint.objects.all().delete()
        self.stdout.write(self.style.MIGRATE_LABEL(f'Deleted Fingerprints: {deleted}'))



from django.core.management.base import BaseCommand

from djspoofer.models import H2Fingerprint
from djspoofer import utils


class Command(BaseCommand):
    help = 'Save H2 Hash'

    def add_arguments(self, parser):
        parser.add_argument(
            "--hash",
            required=True,
            type=str,
            help="H2 Fingerprint Hash",
        )
        parser.add_argument(
            "--user-agent",
            required=True,
            type=str,
            help="User Agent",
        )
        parser.add_argument(
            "--browser-min-major-version",
            required=False,
            type=int,
            help="Browser Minimum Major Version",
        )
        parser.add_argument(
            "--browser-max-major-version",
            required=False,
            type=int,
            help="Browser Minimum Major Version",
        )

    def handle(self, *args, **kwargs):
        try:
            h2_fingerprint = self.create_h2_fingerprint(kwargs)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully saved H2 Fingerprint: {h2_fingerprint}'))

    @staticmethod
    def create_h2_fingerprint(kwargs):
        return utils.h2_hash_to_h2_fingerprint(
            user_agent=kwargs['user_agent'],
            h2_hash=kwargs['hash'],
            browser_min_major_version=kwargs.get('browser_min_major_version'),
            browser_max_major_version=kwargs.get('browser_max_major_version')
        )



import json
from importlib.resources import open_text
from unittest import mock

import httpx
from django.test import TestCase
from httpx import Request, Response, codes

from djspoofer.remote.proxyrack import proxyrack_api, exceptions


class BaseTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.mocked_sleep = mock.patch('time.sleep', return_value=None).start()


class ActiveConnectionsTests(BaseTestCase):
    """
        Active Connections Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.proxyrack.resources', 'active_conns.json') as active_conns_json:
            cls.r_data = json.loads(active_conns_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.r_data
        )

        r_active_connections = proxyrack_api.active_connections(
            mock_client,
        )
        self.assertEquals(len(r_active_connections.connections), 2)

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.r_data
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.active_connections(
                mock_client,
            )


class APIKeyTests(BaseTestCase):
    """
        API Keys Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.proxyrack.resources', 'passwords.json') as passwords_json:
            cls.r_data = json.loads(passwords_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.post.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.r_data
        )

        r_temp = proxyrack_api.generate_temp_api_key(
            mock_client,
            expiration_seconds=60
        )
        self.assertEquals(r_temp.api_key, 'temp-bf3702-be83a4-0bbfc1-be7f58-62cfff')

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.post.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.r_data
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.generate_temp_api_key(
                mock_client,
                expiration_seconds=60
            )


class ProxyCheckTests(BaseTestCase):
    """
        Test Proxy Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception

    @mock.patch.object(httpx, 'get')
    def test_ok(self, mock_get):
        mock_get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
        )

        is_valid_proxy = proxyrack_api.is_valid_proxy(proxies={'http://': 'http://example.com'})
        self.assertTrue(is_valid_proxy)
        self.assertEquals(mock_get.call_count, 1)

    @mock.patch.object(httpx, 'get')
    def test_407(self, mock_get):
        mock_get.return_value = Response(
            request=self.request,
            status_code=407,
        )

        self.assertFalse(proxyrack_api.is_valid_proxy(proxies={'http://': 'http://example.com'}))


class StatsTests(BaseTestCase):
    """
        Stats Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.proxyrack.resources', 'stats.json') as stats_json:
            cls.r_data = json.loads(stats_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.r_data
        )

        r_stats = proxyrack_api.stats(
            mock_client,
        )
        self.assertEquals(r_stats.thread_limit, 10000)

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.r_data
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.stats(
                mock_client,
            )


class IspsTests(BaseTestCase):
    """
        Isps Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.proxyrack.resources', 'us_isps.json') as isps_json:
            cls.r_data = json.loads(isps_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.r_data
        )

        r_isps = proxyrack_api.isps(
            mock_client,
            country='US'
        )
        self.assertEquals(len(r_isps.isps), 54)

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.r_data
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.isps(
                mock_client,
                country='US'
            )


class CountriesTests(BaseTestCase):
    """
        Countries Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.proxyrack.resources', 'countries.json') as countries_json:
            cls.r_data = json.loads(countries_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.r_data
        )

        r_countries = proxyrack_api.countries(
            mock_client,
            country='US'
        )
        self.assertEquals(len(r_countries.countries), 205)

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.r_data
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.countries(
                mock_client,
                country='US'
            )


class CitiesTests(BaseTestCase):
    """
        Cities Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.proxyrack.resources', 'cities.json') as cities_json:
            cls.r_data = json.loads(cities_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.r_data
        )

        r_cities = proxyrack_api.cities(
            mock_client,
            country='US'
        )
        self.assertEquals(len(r_cities.cities), 82)

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.r_data
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.cities(
                mock_client,
                country='US'
            )


class CountryIPCountTests(BaseTestCase):
    """
        Cities Tests
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            text='152'
        )

        r_country_ip_count = proxyrack_api.country_ip_count(
            mock_client,
            country='US'
        )
        self.assertEquals(r_country_ip_count, '152')

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            text='152'
        )

        with self.assertRaises(exceptions.ProxyRackError):
            proxyrack_api.country_ip_count(
                mock_client,
                country='US'
            )



from django.test import TestCase

from djspoofer.remote.proxyrack import const, utils


class UtilTests(TestCase):
    """
    Utility Tests
    """

    def test_proxy_builder(self):
        proxy_builder = utils.ProxyBuilder(
            username='proxyman123',
            password='goodpw567',
            netloc='megaproxy.rotating.proxyrack.net:10000',
            country='US',
            city='Seattle,New York,Los Angeles',
            isp='Verizon,Ipxo Limited',
            refreshMinutes=10,
            osName=const.ProxyOs.LINUX,
            session='13ac97fe-0f26-45ff-aeb9-2801400326ec',
            proxyIp='184.53.48.165',
            missingKey=''
        )

        self.assertEquals(
            proxy_builder.http_url,
            ('http://proxyman123;country=US;city=Seattle,NewYork,LosAngeles;isp=Verizon,IpxoLimited;refreshMinutes=10;'
             'osName=Linux;session=13ac97fe-0f26-45ff-aeb9-2801400326ec;proxyIp=184.53.48.165:'
             'goodpw567@megaproxy.rotating.proxyrack.net:10000')
        )

    def test_proxy_weighted_country(self):
        country = utils.proxy_weighted_country()
        self.assertIsNotNone(country)

    def test_proxy_weighted_isp(self):
        isp = utils.proxy_weighted_isp()
        self.assertIsNotNone(isp)



import json
from importlib.resources import open_text
from unittest import mock

from django.test import TestCase
from httpx import Request, Response, codes

from djspoofer import clients, utils
from djspoofer.models import Fingerprint, DeviceFingerprint, H2Fingerprint, IP, Proxy, Geolocation
from djspoofer.remote.proxyrack import proxyrack_api


class DesktopChromeClientTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.proxy = Proxy.objects.create_rotating_proxy(
            url='test123:5000',
        )
        user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36'
        device_fingerprint_data = {
            'device_category': 'desktop',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': user_agent,
            'viewport_height': 768,
            'viewport_width': 1024
        }
        h2_fingerprint_data = {
            'os': 'Windows',
            'browser': 'Chrome',
            'browser_min_major_version': 95,
            'browser_max_major_version': 100,
            'header_table_size': 65536,
            'enable_push': True,
            'max_concurrent_streams': 1000,
            'initial_window_size': 6291456,
            'max_frame_size': 16384,
            'max_header_list_size': 262144,
            'psuedo_header_order': 'm,a,s,p',
            'window_update_increment': 15663105,
            'priority_stream_id': 1,
            'priority_exclusive': True,
            'priority_depends_on_id': 0,
            'priority_weight': 256
        }
        cls.geo_location_data = {
            'city': 'Los Angeles',
            'country': 'US',
            'isp': 'Spectrum',
        }
        cls.ip_data = {
            'city': 'Dallas',
            'country': 'US',
            'isp': 'Spectrum',
            'address': '194.60.86.250',
        }
        cls.fingerprint = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**device_fingerprint_data),
            _h2_fingerprint=H2Fingerprint.objects.create(**h2_fingerprint_data),
        )
        with open_text('djspoofer.tests.proxyrack.resources', 'stats.json') as stats_json:
            cls.r_stats_data = proxyrack_api.StatsResponse(json.loads(stats_json.read()))

    @mock.patch.object(proxyrack_api, 'stats')
    @mock.patch.object(proxyrack_api, 'is_valid_proxy')
    @mock.patch.object(clients.DesktopChromeClient, '_send_handling_auth')
    def test_ok(self, mock_sd_send, mock_is_valid_proxy, mock_stats):
        mock_sd_send.return_value = Response(
            request=Request(url='', method=''),
            status_code=codes.OK,
            text='ok'
        )
        mock_is_valid_proxy.return_value = True
        mock_stats.return_value = self.r_stats_data

        self.fingerprint.add_ip(IP.objects.create(**self.ip_data))

        with clients.DesktopChromeClient(fingerprint=self.fingerprint) as chrome_client:
            chrome_client.get('http://example.com')
            self.assertEquals(mock_sd_send.call_count, 1)
            self.assertEquals(
                chrome_client.sec_ch_ua,
                '" Not;A Brand";v="99", "Google Chrome";v="99", "Chromium";v="99"'
            )
            self.assertEquals(chrome_client.sec_ch_ua_mobile, '?0')
            self.assertEquals(chrome_client.sec_ch_ua_platform, '"Windows"')

    @mock.patch.object(proxyrack_api, 'stats')
    @mock.patch.object(proxyrack_api, 'is_valid_proxy')
    @mock.patch.object(clients.DesktopChromeClient, '_send_handling_auth')
    def test_fingerprint_with_geolocation_no_ips(self, mock_sd_send, mock_is_valid_proxy, mock_stats):
        mock_sd_send.return_value = Response(
            request=Request(url='', method=''),
            status_code=codes.OK,
            text='ok'
        )
        mock_is_valid_proxy.return_value = True
        mock_stats.return_value = self.r_stats_data

        self.fingerprint.set_geolocation(Geolocation.objects.create(**self.geo_location_data))

        with clients.DesktopChromeClient(fingerprint=self.fingerprint) as chrome_client:
            chrome_client.get('http://example.com')
            self.assertEquals(mock_sd_send.call_count, 1)
            self.assertEquals(
                chrome_client.sec_ch_ua,
                '" Not;A Brand";v="99", "Google Chrome";v="99", "Chromium";v="99"'
            )
            self.assertEquals(chrome_client.sec_ch_ua_mobile, '?0')
            self.assertEquals(chrome_client.sec_ch_ua_platform, '"Windows"')


class DesktopFirefoxClientTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.proxy = Proxy.objects.create_rotating_proxy(
            url='test123:5000',
        )
        user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36'
        device_fingerprint_data = {
            'device_category': 'desktop',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': user_agent,
            'viewport_height': 768,
            'viewport_width': 1024
        }
        h2_fingerprint_data = {
            'os': 'Windows',
            'browser': 'Chrome',
            'browser_min_major_version': 95,
            'browser_max_major_version': 100,
            'header_table_size': 65536,
            'enable_push': True,
            'max_concurrent_streams': 1000,
            'initial_window_size': 6291456,
            'max_frame_size': 16384,
            'max_header_list_size': 262144,
            'psuedo_header_order': 'm,a,s,p',
            'window_update_increment': 15663105,
            'priority_stream_id': 1,
            'priority_exclusive': True,
            'priority_depends_on_id': 0,
            'priority_weight': 256
        }
        cls.fingerprint = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**device_fingerprint_data),
            _h2_fingerprint=H2Fingerprint.objects.create(**h2_fingerprint_data)
        )
        with open_text('djspoofer.tests.proxyrack.resources', 'stats.json') as stats_json:
            cls.r_stats_data = proxyrack_api.StatsResponse(json.loads(stats_json.read()))

    @mock.patch.object(proxyrack_api, 'stats')
    @mock.patch.object(proxyrack_api, 'is_valid_proxy')
    @mock.patch.object(clients.DesktopFirefoxClient, '_send_handling_auth')
    def test_ok(self, mock_sd_send, mock_is_valid_proxy, mock_stats):
        mock_sd_send.return_value = Response(
            request=Request(url='', method=''),
            status_code=codes.OK,
            text='ok'
        )
        mock_is_valid_proxy.return_value = True
        mock_stats.return_value = self.r_stats_data

        with clients.DesktopFirefoxClient(fingerprint=self.fingerprint) as sd_client:
            sd_client.get('http://example.com')
            self.assertEquals(mock_sd_send.call_count, 1)





from django.core.management.base import BaseCommand
from djstarter import utils

from djspoofer.clients import DesktopChromeClient
from djspoofer.remote.h2fingerprint import h2fingerprint_api


class Command(BaseCommand):
    help = 'Get H2 Fingerprint'

    def handle(self, *args, **kwargs):
        try:
            with DesktopChromeClient() as client:
                r_h2 = h2fingerprint_api.get_h2_fingerprint(client)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(utils.pretty_dict(r_h2))
            self.stdout.write(self.style.MIGRATE_LABEL('Finished getting H2 Fingerprint'))



import logging
import random
import uuid

from django.conf import settings
from djstarter import decorators
from httpx import Client

from djspoofer import backends, exceptions, utils
from djspoofer.models import IP, Proxy
from djspoofer.remote.proxyrack import proxyrack_api, utils as pr_utils

logger = logging.getLogger(__name__)


class ProxyRackProxyBackend(backends.ProxyBackend):
    def get_proxy_url(self, fingerprint):
        for ip in fingerprint.get_last_n_ips(count=3):
            proxy_url = self._build_proxy_url(proxyIp=ip.address)
            if self._is_valid_proxy(proxies=utils.proxy_dict(proxy_url)):
                logger.info(f'Found valid IP Fingerprint: {ip}')
                return proxy_url
        else:
            logger.info(f'{fingerprint}. No valid IP Fingerprints found. ')
            return self._new_proxy_url(fingerprint)   # Generate if no valid IP Fingerprints

    @decorators.retry(retry_exceptions=exceptions.ProxyConnectionFailed, tries=3)
    def _new_proxy_url(self, fingerprint):
        logger.info(f'{fingerprint}. Generating new IP Fingerprint. ')
        proxy_url = self._test_proxy_url(fingerprint)
        proxies = utils.proxy_dict(proxy_url)
        if self._is_valid_proxy(proxies=proxies):
            self._create_ip_fingerprint(fingerprint, proxies)
            return proxy_url
        raise exceptions.ProxyConnectionFailed(f'{fingerprint}. Failed to get a new valid proxy')

    @staticmethod
    def _is_valid_proxy(proxies):
        return proxyrack_api.is_valid_proxy(proxies)

    @staticmethod
    def _create_ip_fingerprint(fingerprint, proxies):
        with Client(proxies=proxies) as client:
            r_stats = proxyrack_api.stats(client)
        ip_fingerprint = IP.objects.create(
            city=r_stats.ipinfo.city,
            country=r_stats.ipinfo.country,
            isp=r_stats.ipinfo.isp,
            address=r_stats.ipinfo.ip,
            fingerprint=fingerprint
        )
        fingerprint.add_ip(ip_fingerprint)
        logger.info(f'{ip_fingerprint}. Successfully created new ip fingerprint')

    def _test_proxy_url(self, fingerprint):
        geolocation = fingerprint.geolocation
        logger.info(f'{fingerprint}. Using Geolocation: {geolocation}')
        return self._build_proxy_url(
            osName=fingerprint.device_fingerprint.os,
            country=getattr(geolocation, 'country', self._weighted_proxy_country()),
            city=getattr(geolocation, 'city', None),
            # isp=getattr(geolocation, 'isp', None),    # Appears to be too restrictive
        )

    @staticmethod
    def _build_proxy_url(**kwargs):
        return pr_utils.ProxyBuilder(
            username=settings.PROXY_USERNAME,
            password=settings.PROXY_PASSWORD,
            netloc=Proxy.objects.get_rotating_proxy().url,
            timeoutSeconds=60,
            session=str(uuid.uuid4()),
            **kwargs
        ).http_url

    @staticmethod
    def _weighted_proxy_country():
        countries, weights = zip(*settings.PROXYRACK_COUNTRY_WEIGHTS)
        return random.choices(population=countries, weights=weights, k=1)[0]



import collections
import logging

import h2
from h2.settings import Settings, SettingCodes
from httpcore._models import Request
from httpcore._sync import http2

from djspoofer import exceptions
from djspoofer.models import H2Fingerprint

logger = logging.getLogger(__name__)

H2_FINGERPRINT_HEADER = b'h2-fingerprint-id'


def _send_connection_init(self, request: Request) -> None:
    """
        ** Monkey Patched in apps.py **
        The HTTP/2 connection requires some initial setup before we can start
        using individual request/response streams on it.
    """
    # Need to set these manually here instead of manipulating via
    # __setitem__() otherwise the H2Connection will emit SettingsUpdate
    # frames in addition to sending the undesired defaults.
    h2_frame_fingerprint = _get_h2_fingerprint(request)
    self._h2_state.local_settings = build_h2_settings(h2_frame_fingerprint)

    self._h2_state.get_next_available_stream_id = lambda: h2_frame_fingerprint.priority_stream_id
    self._h2_state.initiate_connection()
    self._h2_state.increment_flow_control_window(h2_frame_fingerprint.window_update_increment)

    self._write_outgoing_data(request)


def _send_request_headers(self, request: Request, stream_id: int) -> None:
    """
        ** Monkey Patched in apps.py **
    """
    end_stream = not http2.has_body_headers(request)

    h2_fingerprint = _get_h2_fingerprint(request)
    headers = _get_psuedo_headers(request, h2_fingerprint=h2_fingerprint) + [
        (k.lower(), v)
        for k, v in request.headers
        if k.lower()
        not in (
            b"host",
            b"transfer-encoding",
            H2_FINGERPRINT_HEADER
        )
    ]

    self._h2_state.send_headers(
        stream_id,
        headers,
        end_stream=end_stream,
        priority_weight=h2_fingerprint.priority_weight,
        priority_depends_on=h2_fingerprint.priority_depends_on_id,
        priority_exclusive=h2_fingerprint.priority_exclusive
    )
    self._h2_state.increment_flow_control_window(h2_fingerprint.window_update_increment, stream_id=stream_id)
    self._write_outgoing_data(request)


def _get_psuedo_headers(request, h2_fingerprint):
    header_map = {
        'm': (b":method", request.method),
        'a': (b":authority", _get_authority(request)),
        's': (b":scheme", request.url.scheme),
        'p': (b":path", request.url.target),
    }
    return [header_map[k] for k in h2_fingerprint.psuedo_header_order.split(',')]


def _get_authority(request):
    """
        In HTTP/2 the ':authority' pseudo-header is used instead of 'Host'.
        In order to gracefully handle HTTP/1.1 and HTTP/2 we always require
        HTTP/1.1 style headers, and map them appropriately if we end up on
        an HTTP/2 connection.
    """
    return [v for k, v in request.headers if k.lower() == b"host"][0]


def _get_h2_fingerprint(request):
    for i, (h_key, h_val) in enumerate(request.headers):
        if h_key == H2_FINGERPRINT_HEADER:
            return H2Fingerprint.objects.get_by_oid(str(h_val, 'utf-8'))
    raise exceptions.DJSpooferError(f'Header "{H2_FINGERPRINT_HEADER}" missing')


def build_h2_settings(h2_settings_fingerprint):
    h2_fp = h2_settings_fingerprint
    initial_values = {
        SettingCodes.HEADER_TABLE_SIZE: h2_fp.header_table_size,                            # 0x01 (Required)
        SettingCodes.ENABLE_PUSH: int(h2_fp.enable_push) if h2_fp.enable_push else None,    # 0x02 (Required)
        SettingCodes.MAX_CONCURRENT_STREAMS: h2_fp.max_concurrent_streams,                  # 0x03 (Optional)
        SettingCodes.INITIAL_WINDOW_SIZE: h2_fp.initial_window_size,                        # 0x04 (Required)
        SettingCodes.MAX_FRAME_SIZE: h2_fp.max_frame_size,                                  # 0x05 (Required)
        SettingCodes.MAX_HEADER_LIST_SIZE: h2_fp.max_header_list_size,                      # 0x06 (Optional)
    }
    initial_values = {k: v for k, v in initial_values.items() if v}
    return H2Settings(initial_values=initial_values)


class H2Settings(h2.settings.Settings):
    """
        Allows for setting the settings value in any particular order.
        There is no validation of settings since validation throws errors for missing or invalid values
        Use with caution!
    """
    def __init__(self, initial_values=None):
        super().__init__()
        self._settings = {k: collections.deque([v]) for k, v in initial_values.items()}



import logging

from djstarter import decorators
from httpx import Client

from djspoofer import utils as s_utils
from djspoofer.remote.intoli import intoli_api, const
from djspoofer.remote.intoli.exceptions import IntoliError
from djspoofer.models import IntoliFingerprint

logger = logging.getLogger(__name__)


@decorators.db_conn_close
def get_profiles(*args, **kwargs):
    GetProfiles(*args, **kwargs).start()


class GetProfiles:
    def __init__(self, *args, **kwargs):
        self.max_profiles = kwargs.get('max_profiles', const.MAX_PROFILES)
        self.desktop_only = kwargs.get('desktop_only')
        self.os_list = kwargs.get('os_list')

    def start(self):
        with Client() as client:
            r_profiles = intoli_api.get_profiles(client)

        old_oids = list(IntoliFingerprint.objects.all_oids())
        profiles = self.build_profiles(r_profiles)

        try:
            new_profiles = IntoliFingerprint.objects.bulk_create(profiles)
        except Exception as e:
            raise IntoliError(info=f'Error adding user agents: {str(e)}')
        else:
            logger.info(f'Max Profiles: {self.max_profiles}')
            logger.info(f'Created New Intoli Profiles: {len(new_profiles)}')
            logger.info(f'Deleted Old Intoli Profiles: {IntoliFingerprint.objects.bulk_delete(oids=old_oids)[0]}')

    def build_profiles(self, r_profiles):
        new_profiles = list()
        for profile in self.filtered_profiles(r_profiles):
            ua_parser = s_utils.UserAgentParser(profile.user_agent)
            temp_profile = IntoliFingerprint(
                browser=ua_parser.browser,
                browser_major_version=ua_parser.browser_major_version,
                os=ua_parser.os,
                device_category=profile.device_category,
                platform=profile.platform,
                screen_height=profile.screen_height,
                screen_width=profile.screen_width,
                user_agent=profile.user_agent,
                viewport_height=profile.viewport_height,
                viewport_width=profile.viewport_width,
                weight=profile.weight,
            )
            new_profiles.append(temp_profile)
        return new_profiles

    def filtered_profiles(self, r_profiles):
        if self.desktop_only:
            profiles = r_profiles.valid_desktop_profiles
        else:
            profiles = r_profiles.valid_profiles
        if self.os_list:
            profiles = [p for p in profiles if p.os in self.os_list]
        return profiles[:self.max_profiles]



from django.conf import settings
from django.core.management.base import BaseCommand
from djstarter import utils

from djspoofer.clients import DesktopChromeClient
from djspoofer.remote.h2fingerprint import h2fingerprint_api
from djspoofer.remote.howsmyssl import howsmyssl_api
from djspoofer.remote.incolumitas import incolumitas_api, incolumitas_tcpip_api, incolumitas_tls_api
from djspoofer.remote.ja3er import ja3er_api
from djspoofer.remote.proxyrack import utils as pr_utils


class Command(BaseCommand):
    help = 'Get All Chrome Fingerprints'

    def add_arguments(self, parser):
        parser.add_argument(
            "--proxy-url",
            required=True,
            type=str,
            help="Set the proxy url",
        )
        parser.add_argument(
            "--proxy-args",
            required=False,
            nargs='*',
            help="Set the proxy password",
        )

    def handle(self, *args, **kwargs):
        try:
            with DesktopChromeClient() as client:
                self.show_ja3er_details(client)
                self.show_ssl_check(client)
                self.show_ip_fingerprint(client)
                self.show_tcpip_fingerprint(client)
                self.show_tls_fingerprint(client)
                self.show_h2_fingerprint(client)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL('Finished getting All Fingerprints'))

    def proxy_http_url(self, proxy_url, proxy_args):
        proxy_builder = pr_utils.ProxyBuilder(
            netloc=proxy_url,
            password=settings.PROXY_PASSWORD,
            username=settings.PROXY_USERNAME,
            **self.proxy_options(proxy_args),
        )
        return proxy_builder.http_url

    def show_ip_fingerprint(self, client):
        r_tls = incolumitas_api.ip_fingerprint(client)
        self.stdout.write(utils.eye_catcher_line('IP Fingerprint'))
        self.stdout.write(utils.pretty_dict(r_tls))

    def show_tcpip_fingerprint(self, client):
        r_tcpip = incolumitas_tcpip_api.tcpip_fingerprint(client)
        self.stdout.write(utils.eye_catcher_line('TCP/IP Fingerprint'))
        self.stdout.write(utils.pretty_dict(r_tcpip))

    def show_tls_fingerprint(self, client):
        try:
            r_tls = incolumitas_tls_api.tls_fingerprint(client)
            self.stdout.write(utils.eye_catcher_line('TLS Fingerprint'))
            self.stdout.write(utils.pretty_dict(r_tls))
        except Exception as e:
            self.stdout.write(utils.eye_catcher_line(f'TLS Fingerprint Failed: {str(e)}'))

    def show_ja3er_details(self, client):
        r_json = ja3er_api.details(client)
        self.stdout.write(utils.eye_catcher_line('JA3 Details'))
        self.stdout.write(utils.pretty_dict(r_json))

        r_search = ja3er_api.search(client, ja3_hash=r_json.ja3_hash)
        self.stdout.write(utils.eye_catcher_line('JA3 Hash Search'))
        self.stdout.write(utils.pretty_dict(vars(r_search)))

    def show_ssl_check(self, client):
        self.stdout.write(utils.eye_catcher_line('SSL Check'))
        try:
            r_check = howsmyssl_api.ssl_check(client)
            self.stdout.write(utils.pretty_dict(vars(r_check)))
        except Exception as e:
            self.stdout.write(utils.eye_catcher_line(f'SSL Check Failed: {str(e)}'))

    def show_h2_fingerprint(self, client):
        self.stdout.write(utils.eye_catcher_line('H2 Fingerprint Check'))
        try:
            r_h2 = h2fingerprint_api.get_h2_fingerprint(client)
            self.stdout.write(utils.pretty_dict(vars(r_h2)))
        except Exception as e:
            self.stdout.write(utils.eye_catcher_line(f'H2 Fingerprint Check Failed: {str(e)}'))

    @staticmethod
    def proxy_options(proxy_args):
        return {args.split('=')[0]: args.split('=')[1] for args in proxy_args}



from django.test import TestCase

from djspoofer.remote.intoli import exceptions
from djspoofer.models import IntoliFingerprint


class ProfileManagerTests(TestCase):
    """
    ProfileManager Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.profile_data = {
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36',
            'viewport_height': 768,
            'viewport_width': 1024,
            'weight': .005,
        }

    def test_all_oids(self):
        profile = IntoliFingerprint.objects.create(device_category='desktop', **self.profile_data)

        self.assertEquals(list(IntoliFingerprint.objects.all_oids()), [profile.oid])

    def test_all_user_agents(self):
        IntoliFingerprint.objects.create(**self.profile_data)

        new_data = self.profile_data.copy()
        new_data['user_agent'] = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.54 Safari/537.36'
        IntoliFingerprint.objects.create(**new_data)

        user_agents = IntoliFingerprint.objects.all_user_agents()

        self.assertListEqual(
            list(user_agents),
            [
                'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36',
                'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.54 Safari/537.36'
            ]
        )

    def test_random_desktop_chrome_profile(self):
        with self.assertRaises(exceptions.IntoliError):
            IntoliFingerprint.objects.random_desktop()

        IntoliFingerprint.objects.create(browser='Chrome', device_category='desktop', os='Windows', **self.profile_data)

        profile = IntoliFingerprint.objects.random_desktop()

        self.assertEquals(profile.user_agent, 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36')

    def test_random_mobile_profile(self):
        with self.assertRaises(exceptions.IntoliError):
            IntoliFingerprint.objects.random_mobile()

        IntoliFingerprint.objects.create(device_category='mobile', **self.profile_data)

        profile = IntoliFingerprint.objects.random_mobile()

        self.assertEquals(profile.user_agent, 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36')

    def test_weighted_desktop_chrome_user_agent(self):
        with self.assertRaises(exceptions.IntoliError):
            IntoliFingerprint.objects.weighted_n_desktop()

        IntoliFingerprint.objects.create(browser='Chrome', device_category='desktop', os='Windows', **self.profile_data)

        profile = IntoliFingerprint.objects.weighted_n_desktop()[0]

        self.assertEquals(profile.user_agent, 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36')

    def test_weighted_mobile_user_agent(self):
        with self.assertRaises(exceptions.IntoliError):
            IntoliFingerprint.objects.weighted_n_mobile()

        IntoliFingerprint.objects.create(device_category='mobile', **self.profile_data)

        profile = IntoliFingerprint.objects.weighted_n_mobile()[0]

        self.assertEquals(profile.user_agent, 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36')

    def test_bulk_delete(self):
        profile = IntoliFingerprint.objects.create(device_category='desktop', os='Linux', **self.profile_data)

        IntoliFingerprint.objects.bulk_delete(oids=[profile.oid])

        with self.assertRaises(IntoliFingerprint.DoesNotExist):
            profile.refresh_from_db()



import gzip
import json
import logging
from io import BytesIO

from django.conf import settings
from djstarter import decorators

from djspoofer import utils
from .exceptions import IntoliError
from . import const

logger = logging.getLogger(__name__)

BASE_URL = settings.INTOLI_API_BASE_URL


@decorators.wrap_exceptions(raise_as=IntoliError)
def get_profiles(client):
    url = f'{BASE_URL}/intoli/user-agents/master/src/user-agents.json.gz'

    params = {
        'format': 'json',
    }

    with client.stream('GET', url, params=params) as response:
        json_io = BytesIO()
        try:
            for chunk in response.iter_bytes(chunk_size=8192):
                json_io.write(chunk)
                json_io.flush()
            json_io.seek(0)
            with gzip.GzipFile(fileobj=json_io, mode='rb') as gz_file:
                return GetProfilesResponse(json.load(gz_file))
        finally:
            json_io.close()


class GetProfilesResponse:
    def __init__(self, json_data):
        self.profiles = sorted([IntoliProfile(profile) for profile in json_data], key=lambda p: p.weight, reverse=True)

    @property
    def valid_mobile_profiles(self):
        return [p for p in self.valid_profiles if p.is_mobile]

    @property
    def valid_desktop_profiles(self):
        return [p for p in self.valid_profiles if p.is_desktop]

    @property
    def valid_profiles(self):
        return [p for p in self.profiles if p.is_valid_profile]


class IntoliProfile:

    def __init__(self, data):
        self.user_agent = data.get('userAgent').strip(const.USER_AGENT_BAD_CHARS)
        ua_parser = utils.UserAgentParser(self.user_agent)
        self.browser = ua_parser.browser
        self.os = ua_parser.os
        self.device_category = data.get('deviceCategory')
        self.platform = data.get('platform')
        self.screen_height = data.get('screenHeight') or 1080
        self.screen_width = data.get('screenWidth') or 1920
        self.viewport_height = data.get('viewportHeight') or 920
        self.viewport_width = data.get('viewportWidth') or 1415
        self.weight = data.get('weight')

    @property
    def is_mobile(self):
        return self.device_category == 'mobile'

    @property
    def is_desktop(self):
        return self.device_category == 'desktop'

    @property
    def is_valid_profile(self):
        return all([
            self.is_valid_user_agent,
            not self.is_user_agent_blacklisted,
            self.is_valid_device_category,
            self.is_valid_platform,
        ])

    @property
    def is_valid_user_agent(self):
        return const.USER_AGENT_LEN_RANGE[0] <= len(self.user_agent) <= const.USER_AGENT_LEN_RANGE[1]

    @property
    def is_user_agent_blacklisted(self):
        return any(x.lower() in self.user_agent.lower() for x in const.USER_AGENT_BLACKLIST)

    @property
    def is_valid_device_category(self):
        return const.DEVICE_CATEGORY_LEN_RANGE[0] <= len(self.device_category) <= const.DEVICE_CATEGORY_LEN_RANGE[1]

    @property
    def is_valid_platform(self):
        return const.PLATFORM_LEN_RANGE[0] <= len(self.platform) <= const.PLATFORM_LEN_RANGE[1]



from django.test import TestCase

from djspoofer import utils


class UtilTests(TestCase):
    """
    Utility Tests
    """

    def test_fake_profile(self):
        old_profile = utils.FakeProfile()
        profile = utils.FakeProfile()
        self.assertNotEquals(old_profile, profile)

        self.assertIn(profile.gender, ['M', 'F'])
        self.assertIn(profile.full_gender, ['MALE', 'FEMALE'])
        self.assertEquals(profile.full_name, f'{profile.first_name} {profile.last_name}')

        dob = profile.dob
        self.assertEquals(profile.dob_yyyymmdd, f'{dob.year}-{dob.month:02}-{dob.day:02}')
        self.assertTrue(profile.us_phone_number.startswith('+1'))
        self.assertEquals(len(profile.us_phone_number), 12)

    def test_ua_parser(self):
        user_agent = ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                      'Chrome/99.0.4844.82 Safari/537.36')
        ua_parser = utils.UserAgentParser(user_agent=user_agent)

        self.assertEquals(
            str(ua_parser),
            ("UserAgentParser -> {'user_agent': {'family': 'Chrome', 'major': '99', 'minor': '0', 'patch': '4844'}, 'os': {'family': 'Windows', 'major': '10', 'minor': None, 'patch': None, 'patch_minor': None}, 'device': {'family': 'Other', 'brand': None, 'model': None}, 'string': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36'}")
        )

        user_agent = ua_parser._user_agent
        self.assertEquals(user_agent.family, 'Chrome')
        self.assertEquals(user_agent.major, '99')
        self.assertEquals(user_agent.minor, '0')
        self.assertEquals(user_agent.patch, '4844')

        os = ua_parser._os
        self.assertEquals(os.family, 'Windows')
        self.assertEquals(os.major, '10')
        self.assertIsNone(os.minor)
        self.assertIsNone(os.patch)

        self.assertEquals(ua_parser.browser, 'Chrome')
        self.assertEquals(ua_parser.browser_major_version, '99')
        self.assertEquals(ua_parser.os, 'Windows')

    def test_h2_fingerprint_parser(self):
        parser = utils.H2HashParser('1:65536;2:1;3:1000;4:6291456;5:16384;6:262144|15663105|1:1:0:256|m,a,s,p')

        settings_frame = parser.settings_frame
        self.assertEquals(settings_frame.header_table_size, 65536)
        self.assertEquals(settings_frame.push_enabled, True)
        self.assertEquals(settings_frame.max_concurrent_streams, 1000)
        self.assertEquals(settings_frame.initial_window_size, 6291456)
        self.assertEquals(settings_frame.max_frame_size, 16384)
        self.assertEquals(settings_frame.max_header_list_size, 262144)

        windows_frame = parser.window_frame
        self.assertEquals(windows_frame, 15663105)

        priority_frame = parser.priority_frame
        self.assertEquals(priority_frame.stream_id, 1)
        self.assertEquals(priority_frame.is_exclusive, 1)
        self.assertEquals(priority_frame.depends_on_id, 0)
        self.assertEquals(priority_frame.weight, 256)

    def test_h2_hash(self):
        h2_hash = '1:65536;2:1;3:1000;4:6291456;5:16384;6:262144|15663105|1:1:0:256|m,a,s,p'
        h2_fingerprint = utils.h2_hash_to_h2_fingerprint(
            os='Windows',
            browser='Chrome',
            h2_hash=h2_hash,
            browser_min_major_version=50,
            browser_max_major_version=100
        )
        self.assertEquals(h2_fingerprint.browser, 'Chrome')
        self.assertEquals(h2_fingerprint.os, 'Windows')
        self.assertEquals(h2_fingerprint.initial_window_size, 6291456)
        self.assertEquals(h2_fingerprint.window_update_increment, 15663105)
        self.assertEquals(h2_fingerprint.priority_weight, 256)



import json
from importlib.resources import open_text
from unittest import mock

import httpx
from django.test import TestCase
from djspoofer.remote.h2fingerprint import exceptions, h2fingerprint_api
from httpx import Request, Response, codes


class DetailsTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.h2_fingerprint.resources', 'response.json') as response_json:
            cls.response_json = json.loads(response_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.response_json
        )

        r_response = h2fingerprint_api.get_h2_fingerprint(mock_client)

        self.assertEquals(r_response.fingerprint, '1:65536;3:1000;4:6291456;6:262144|15663105|1:1:0:256|m,a,s,p')
        self.assertEquals(r_response.settings_frame, '1:65536;3:1000;4:6291456;6:262144')
        self.assertEquals(r_response.window_frame, '15663105')
        self.assertEquals(r_response.priority_frame, '1:1:0:256')
        self.assertEquals(r_response.pseudo_headers, 'm,a,s,p')
        self.assertEquals(r_response.user_agent, 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.54 Safari/537.36')

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.response_json
        )

        with self.assertRaises(exceptions.H2Error):
            h2fingerprint_api.get_h2_fingerprint(
                mock_client,
            )



import logging

from django.conf import settings
from djstarter import decorators

from .exceptions import IncolumitasError

logger = logging.getLogger(__name__)

BASE_URL = settings.INCOLUMITAS_API_BASE_URL


@decorators.wrap_exceptions(raise_as=IncolumitasError)
def ip_fingerprint(client, ip_addr=None):
    url = f'{BASE_URL}/datacenter'

    params = dict()

    if ip_addr:
        params['ip'] = ip_addr

    r = client.get(url, params=params)
    r.raise_for_status()
    return IPFingerprintResponse(r.json())


class IPFingerprintResponse:
    def __init__(self, data):
        self.ip = data['ip']
        self.is_datacenter = data['is_datacenter']
        self.elapsed_ms = data['elapsed_ms']



from importlib.resources import open_binary
from unittest import mock

import httpx
from django.test import TestCase
from httpx import Request

from djspoofer.remote.intoli import intoli_api


class GetProfilesTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        with open_binary('djspoofer.tests.intoli.resources', 'user-agents.json.gz') as intoli_gz:

            mock_client.stream.return_value.__enter__.return_value = mock.Mock()
            mock_client.stream.return_value.__enter__.return_value.status_code = 200
            mock_client.stream.return_value.__enter__.return_value.iter_bytes.return_value = [
                intoli_gz.read()
            ]

        r_profiles = intoli_api.get_profiles(
            mock_client,
        )
        self.assertEquals(len(r_profiles.profiles), 5594)
        self.assertEquals(len(r_profiles.valid_profiles), 5515)



from django.test import TestCase

from djspoofer.models import IntoliFingerprint


class ProfileTests(TestCase):
    """
    Profile Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.profile_data = {
            'device_category': 'mobile',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36',
            'viewport_height': 768,
            'viewport_width': 1024,
            'weight': .005,
        }

    def test_user_str(self):
        profile = IntoliFingerprint.objects.create(**self.profile_data)
        self.assertEqual(
            str(profile),
            ('IntoliFingerprint -> user_agent: Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 '
             '(KHTML, like Gecko) Chrome/99.0.4844.82 Safari/537.36, device_category: mobile, platform: US')
        )

    def test_is_desktop(self):
        profile = IntoliFingerprint.objects.create(**self.profile_data)
        self.assertFalse(profile.is_desktop)

    def test_is_mobile(self):
        profile = IntoliFingerprint.objects.create(**self.profile_data)
        self.assertTrue(profile.is_mobile)



class ProxyOs:
    FREE_BSD = 'FreeBSD'
    LINUX = 'Linux'
    # MAC_OS_X = 'Mac OS X' # Does not work reliably
    WINDOWS = 'Windows'


TOP_RESIDENTIAL_US_ISPS = [
    'AT&T U-verse',
    'Comcast Cable',
    'Cox Communications',
    'Hughes Network Systems',
    'Mediacom Cable',
    'Spectrum',
    'Verizon Fios'
]

DEFAULT_PROXYRACK_COUNTRY_WEIGHTS = [
    ('US', .70),
    ('CA', .12),
    ('UK', .12),
    ('AU', .06),
]

DEFAULT_PROXYRACK_ISP_WEIGHTS = [
    ('AT&T U-verse', 1),
    ('Comcast Cable', 1),
    ('Cox Communications', 1),
    ('Hughes Network Systems', 1),
    ('Mediacom Cable', 1),
    ('Spectrum', 1),
    ('Verizon Fios', 1),
]


# Generated by Django 3.2.13 on 2022-05-06 22:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djspoofer', '0002_alter_proxy_url'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fingerprint',
            name='_h2_fingerprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='fingerprints', to='djspoofer.h2fingerprint'),
        ),
        migrations.AlterField(
            model_name='fingerprint',
            name='_tls_fingerprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='fingerprints', to='djspoofer.tlsfingerprint'),
        ),
        migrations.AlterField(
            model_name='fingerprint',
            name='device_fingerprint',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='fingerprints', to='djspoofer.devicefingerprint'),
        ),
        migrations.AlterField(
            model_name='fingerprint',
            name='geolocation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='fingerprints', to='djspoofer.geolocation'),
        ),
    ]



import datetime
import random
import logging

from django.db import models
from django.utils import timezone
from djstarter.models import BaseModel

from . import const, exceptions, managers, utils

logger = logging.getLogger(__name__)


class BaseFingerprint(BaseModel):
    browser = models.CharField(max_length=32)
    browser_min_major_version = models.IntegerField(default=0)
    browser_max_major_version = models.IntegerField(default=999)
    os = models.CharField(max_length=32)

    class Meta:
        abstract = True


class H2Fingerprint(BaseFingerprint):
    objects = managers.H2FingerprintManager()

    header_table_size = models.IntegerField()
    enable_push = models.BooleanField(blank=True, null=True)
    max_concurrent_streams = models.IntegerField(blank=True, null=True)
    initial_window_size = models.IntegerField()
    max_frame_size = models.IntegerField(blank=True, null=True)
    max_header_list_size = models.IntegerField(blank=True, null=True)
    psuedo_header_order = models.TextField()

    window_update_increment = models.IntegerField()

    priority_stream_id = models.IntegerField()
    priority_exclusive = models.BooleanField()
    priority_depends_on_id = models.IntegerField()
    priority_weight = models.IntegerField()

    class Meta:
        db_table = 'djspoofer_h2_fingerprint'
        ordering = ['-created']
        app_label = 'djspoofer'

    def __str__(self):
        return f'H2Fingerprint -> oid: {self.oid}'


class TLSFingerprint(BaseFingerprint):
    objects = managers.TLSFingerprintManager()

    extensions = models.IntegerField()
    ciphers = models.TextField()

    class Meta:
        db_table = 'djspoofer_tls_fingerprint'
        ordering = ['-created']
        app_label = 'djspoofer'

        indexes = [
            models.Index(fields=['browser', ], name='tls_fp_browser'),
        ]

    def generate_chrome_desktop_cipher_str(self):
        grease_cipher = f'TLS_GREASE_IS_THE_WORD_{random.randint(1, 8)}A'
        return ':'.join(
            [grease_cipher] + [c for c in self.shuffled_ciphers(ciphers=const.ChromeDesktopCiphers, start_idx=4)]
        )

    def generate_firefox_desktop_cipher_str(self):
        return ':'.join([c for c in self.shuffled_ciphers(ciphers=const.FirefoxDesktopCiphers, start_idx=3)])

    DESKTOP_CLIENT_CIPHER_MAP = {
        'Chrome': generate_chrome_desktop_cipher_str,
        'Firefox': generate_firefox_desktop_cipher_str
    }

    @staticmethod
    def shuffled_ciphers(ciphers, start_idx=0, min_k=6):
        first_ciphers = ciphers[:start_idx]
        rem_ciphers = ciphers[start_idx:]
        k = random.randint(min_k, len(rem_ciphers))
        return first_ciphers + random.sample(rem_ciphers, k=k)

    @staticmethod
    def random_tls_extension_int(min_k=4):
        k = random.randint(min_k, len(const.TLS_EXTENSIONS))
        ext_val = 0
        for ext in random.sample(const.TLS_EXTENSIONS, k=k):
            ext_val |= ext
        return int(ext_val)

    def save(self, *args, **kwargs):
        if not self.ciphers:
            self.ciphers = self.DESKTOP_CLIENT_CIPHER_MAP[self.browser](self)
        if not self.extensions:
            self.extensions = self.random_tls_extension_int()
        super().save(*args, **kwargs)


class BaseDeviceFingerprint(BaseModel):
    browser = models.CharField(max_length=32)
    browser_major_version = models.IntegerField()
    device_category = models.CharField(max_length=32)
    os = models.CharField(max_length=32)
    platform = models.CharField(max_length=32)
    screen_height = models.IntegerField()
    screen_width = models.IntegerField()
    user_agent = models.TextField()
    viewport_height = models.IntegerField()
    viewport_width = models.IntegerField()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not all([self.browser, self.browser_major_version, self.os]):
            ua_parser = utils.UserAgentParser(self.user_agent)
            self.browser = ua_parser.browser
            self.browser_major_version = ua_parser.browser_major_version
            self.os = ua_parser.os
        super().save(*args, **kwargs)


class DeviceFingerprint(BaseDeviceFingerprint):
    class Meta:
        db_table = 'djspoofer_device_fingerprint'
        app_label = 'djspoofer'

        indexes = [
            models.Index(fields=['browser', ], name='idx_device_fp_browser'),
            models.Index(fields=['device_category', ], name='idx_device_fp_device_category'),
            models.Index(fields=['platform', ], name='idx_device_fp_platform'),
        ]

    def __str__(self):
        return (f'DeviceFingerprint -> user_agent: {self.user_agent}, device_category: {self.device_category}, '
                f'platform: {self.platform}')


class IntoliFingerprint(BaseDeviceFingerprint):
    objects = managers.IntoliFingerprintManager()

    weight = models.DecimalField(max_digits=25, decimal_places=24)

    class Meta:
        db_table = 'djspoofer_intoli_fingerprint'
        ordering = ['-weight']
        app_label = 'djspoofer'

        indexes = [
            models.Index(fields=['browser', ], name='idx_intoli_fp_browser'),
            models.Index(fields=['device_category', ], name='idx_intoli_fp_device_category'),
            models.Index(fields=['os', ], name='idx_intoli_fp_os'),
            models.Index(fields=['platform', ], name='idx_intoli_fp_platform'),
        ]

    def __str__(self):
        return (f'IntoliFingerprint -> user_agent: {self.user_agent}, device_category: {self.device_category}, '
                f'platform: {self.platform}')

    @property
    def is_desktop(self):
        return self.device_category == 'desktop'

    @property
    def is_mobile(self):
        return self.device_category == 'mobile'


class Geolocation(BaseModel):
    objects = managers.GeolocationManager()

    city = models.CharField(max_length=64)
    country = models.CharField(max_length=2)
    isp = models.CharField(max_length=64, blank=True, null=True)

    class Meta:
        db_table = 'djspoofer_geolocation'
        ordering = ['-created']
        app_label = 'djspoofer'

    def __str__(self):
        return f'Geolocation -> city: {self.city}, country: {self.country}, isp: {self.isp}'


class IP(BaseModel):
    objects = managers.IPManager()

    city = models.CharField(max_length=64)
    country = models.CharField(max_length=2)
    isp = models.CharField(max_length=64)
    address = models.GenericIPAddressField()

    class Meta:
        db_table = 'djspoofer_ip'
        ordering = ['-created']
        app_label = 'djspoofer'

        indexes = [
            models.Index(fields=['city', ], name='ip_fp_city'),
            models.Index(fields=['country', ], name='ip_fp_country'),
            models.Index(fields=['isp', ], name='ip_fp_isp'),
        ]

    def __str__(self):
        return f'IP -> address: {self.address}'


class Fingerprint(BaseModel):
    objects = managers.FingerprintManager()

    device_fingerprint = models.ForeignKey(
        to=DeviceFingerprint,
        related_name='fingerprints',
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )

    _h2_fingerprint = models.ForeignKey(
        to=H2Fingerprint,
        related_name='fingerprints',
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )

    _tls_fingerprint = models.ForeignKey(
        to=TLSFingerprint,
        related_name='fingerprints',
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )

    geolocation = models.ForeignKey(
        to=Geolocation,
        related_name='fingerprints',
        on_delete=models.SET_NULL,
        blank=True,
        null=True
    )

    ip = models.ManyToManyField(IP)

    class Meta:
        db_table = 'djspoofer_fingerprint'
        ordering = ['-created']
        app_label = 'djspoofer'

    def __str__(self):
        return f'Fingerprint -> user_agent: {self.device_fingerprint.user_agent}'

    @property
    def h2_fingerprint(self):
        if not self._h2_fingerprint:
            new_h2_fingerprint = H2Fingerprint.objects.get_by_browser_info(
                os=self.device_fingerprint.os,
                browser=self.device_fingerprint.browser,
                browser_major_version=self.device_fingerprint.browser_major_version
            )
            if not new_h2_fingerprint:
                raise exceptions.DJSpooferError('No Available H2 Fingerprints. Did you run the djspoofer_init command?')
            self._h2_fingerprint = new_h2_fingerprint
            self.save()
        return self._h2_fingerprint

    @property
    def tls_fingerprint(self):
        if not self._tls_fingerprint:
            tls_fingerprint = TLSFingerprint.objects.create(browser=self.device_fingerprint.browser)
            if not tls_fingerprint:
                raise exceptions.DJSpooferError('No Available TLS Fingerprints. Did you run the djspoofer_init command?')
            self._tls_fingerprint = tls_fingerprint
            self.save()
        return self._tls_fingerprint

    def set_geolocation(self, geolocation):
        self.geolocation = geolocation
        self.save()

    def get_last_n_ips(self, count=3):
        return self.ip.all().order_by('-created')[:count]

    def add_ip(self, ip):
        self.ip.add(ip)
        if not self.geolocation:
            self.geolocation = Geolocation.objects.create(
                city=ip.city,
                country=ip.country,
                isp=ip.isp,
            )
            logger.info(f'{self}. Now using Geolocation: {self.geolocation}')
        self.save()


class Proxy(BaseModel):
    objects = managers.ProxyManager()

    url = models.TextField(blank=False)
    mode = models.IntegerField(default=const.ProxyModes.GENERAL.value, choices=const.ProxyModes.choices())
    country = models.CharField(max_length=3, blank=True, null=True)
    city = models.CharField(max_length=64, blank=True, null=True)
    last_used = models.DateTimeField(blank=True, null=True)
    used_count = models.IntegerField(default=0)
    cooldown = models.DurationField(default=datetime.timedelta(minutes=10))

    class Meta:
        db_table = 'djspoofer_proxy'
        ordering = ['url']
        app_label = 'djspoofer'

    def __str__(self):
        return f'Proxy -> url: {self.url}, mode: {self.pretty_mode}'

    @property
    def http_url(self):
        return f'http://{self.url}'

    @property
    def https_url(self):
        return f'https://{self.url}'

    @property
    def is_on_cooldown(self):
        if self.last_used:
            return self.last_used > timezone.now() - self.cooldown
        return False

    @property
    def pretty_mode(self):
        return self.get_mode_display()

    def set_last_used(self):
        self.last_used = timezone.now()
        self.used_count += 1
        self.save()



from import_export import resources

from .models import Proxy


class ProxyAdminResource(resources.ModelResource):

    class Meta:
        model = Proxy
        import_id_fields = ('oid',)
        fields = ('oid', 'url', 'mode', 'country', 'city')



import uuid

from django.core.management.base import BaseCommand
from djstarter import utils

from djspoofer import const
from djspoofer.models import Proxy


class Command(BaseCommand):
    help = 'Generate Proxy CSV'

    PROXY_FIELDS = ['oid', 'url', 'mode', 'country', 'city']

    def add_arguments(self, parser):
        parser.add_argument(
            "--url",
            required=True,
            type=str,
            help="Target URL for proxies",
        )

        parser.add_argument(
            "--port-start",
            required=True,
            type=int,
            help="Proxy port start range",
        )
        parser.add_argument(
            "--port-end",
            required=True,
            type=int,
            help="Proxy port end range",
        )
        parser.add_argument(
            "--credentials",
            type=str,
            help="Proxy credentials",
        )
        parser.add_argument(
            "--proxy-mode",
            default=const.ProxyModes.STICKY.value,
            type=int,
            help="Proxy Mode",
        )
        parser.add_argument(
            "--load-proxies",
            type=bool,
            help="Load Proxies Into Database",
        )
        parser.add_argument(
            "--country",
            type=str,
            default='',
            help="Proxy Country",
        )

    def handle(self, *args, **kwargs):
        url = kwargs['url']
        port_start = kwargs['port_start']
        port_end = kwargs['port_end']
        proxy_mode = kwargs['proxy_mode']
        credentials = kwargs['credentials']
        load_proxies = kwargs['load_proxies']
        country = kwargs['country']

        proxies = list()
        self.stdout.write(self.style.MIGRATE_LABEL(','.join(self.PROXY_FIELDS)))
        try:
            for port in range(port_start, port_end):
                columns = [
                    uuid.uuid4(),
                    self.build_proxy_str(url, port=port, credentials=credentials),
                    proxy_mode,
                    country,
                    ''
                ]
                proxies.append(self.build_proxy(columns))
                row = ','.join(str(c) for c in columns)
                self.stdout.write(self.style.MIGRATE_LABEL(row))
            if load_proxies:
                Proxy.objects.bulk_create(proxies)
                self.stdout.write(self.style.MIGRATE_LABEL(utils.eye_catcher_line('Successfully loaded proxies')))

        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(utils.eye_catcher_line('Successfully generated proxy csv')))

    @staticmethod
    def build_proxy(columns):
        return Proxy(
            oid=columns[0],
            url=columns[1],
            mode=columns[2],
            country=columns[3],
            city=columns[4]
        )

    @staticmethod
    def build_proxy_str(url, port, credentials=None):
        proxy_str = f'{url}:{port}'
        if credentials:
            proxy_str = f'{credentials}@{proxy_str}'
        return proxy_str



from ssl import Options

from django.test import TestCase

from djspoofer.models import Geolocation, DeviceFingerprint, H2Fingerprint, TLSFingerprint, Fingerprint, IP, Proxy
from djspoofer.exceptions import DJSpooferError


class FingerprintTests(TestCase):
    """
    Fingerprint Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.device_fingerprint_data = {
            'device_category': 'mobile',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                           'Chrome/99.0.4844.74 Safari/537.36'),
            'viewport_height': 768,
            'viewport_width': 1024,
        }
        cls.ip_fingerprint_data = {
            'city': 'Los Angeles',
            'country': 'US',
            'isp': 'Spectrum',
            'address': '194.60.86.250',
        }
        cls.h2_fingerprint_data = {
            'os': 'Windows',
            'browser': 'Chrome',
            'browser_min_major_version': 95,
            'browser_max_major_version': 100,
            'header_table_size': 65536,
            'enable_push': True,
            'max_concurrent_streams': 1000,
            'initial_window_size': 6291456,
            'max_frame_size': 16384,
            'max_header_list_size': 262144,
            'psuedo_header_order': 'm,a,s,p',
            'window_update_increment': 15663105,
            'priority_stream_id': 1,
            'priority_exclusive': True,
            'priority_depends_on_id': 0,
            'priority_weight': 256
        }

    def test_str(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )
        self.assertEqual(str(fp), f'Fingerprint -> user_agent: {self.device_fingerprint_data["user_agent"]}')

    def test_ok(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )
        self.assertEquals(fp.tls_fingerprint.browser, fp.device_fingerprint.browser)

    def test_set_geolocation(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )
        geolocation = Geolocation.objects.create(city='Los Angeles')
        fp.set_geolocation(geolocation)
        self.assertEquals(fp.tls_fingerprint.browser, fp.device_fingerprint.browser)

    def test_get_first_n_ip_fingerprints(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )
        self.assertEquals(fp.get_last_n_ips(3).count(), 0)

        for _ in range(6):
            fp.add_ip(IP.objects.create(**self.ip_fingerprint_data))

        self.assertEquals(fp.get_last_n_ips(4).count(), 4)

    def test_add_ip_fingerprint(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )

        self.assertIsNone(fp.geolocation)

        ip_fingerprint = IP.objects.create(**self.ip_fingerprint_data)
        fp.add_ip(ip_fingerprint)

        self.assertEquals(fp.ip.all().count(), 1)
        self.assertEquals(fp.geolocation.isp, 'Spectrum')

    def test_h2_fingerprint(self):
        H2Fingerprint.objects.create(**self.h2_fingerprint_data)
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data),
        )
        self.assertIsInstance(fp.h2_fingerprint, H2Fingerprint)

    def test_no_h2_fingerprint(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )
        with self.assertRaises(DJSpooferError):
            _ = fp.h2_fingerprint

    def test_tls_fingerprint(self):
        fp = Fingerprint.objects.create(
            device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        )
        self.assertIsInstance(fp.tls_fingerprint, TLSFingerprint)


class IPTests(TestCase):
    """
    IP Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.ip_data = {
            'city': 'Dallas',
            'country': 'US',
            'isp': 'Spectrum',
            'address': '194.60.86.250',
        }
        cls.fingerprint_data = {
            'browser': 'Chrome',
            'device_category': 'mobile',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                           'Chrome/99.0.4844.74 Safari/537.36'),
            'viewport_height': 768,
            'viewport_width': 1024,
        }

    def test_str(self):
        ip_fingerprint = IP.objects.create(
            **self.ip_data,
        )
        self.assertEqual(str(ip_fingerprint), f'IP -> address: 194.60.86.250')


class H2FingerprintTests(TestCase):
    """
    H2Fingerprint Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.h2_fingerprint_data = {
            'os': 'Windows',
            'browser': 'Chrome',
            'browser_min_major_version': 95,
            'browser_max_major_version': 100,
            'header_table_size': 65536,
            'enable_push': True,
            'max_concurrent_streams': 1000,
            'initial_window_size': 6291456,
            'max_frame_size': 16384,
            'max_header_list_size': 262144,
            'psuedo_header_order': 'm,a,s,p',
            'window_update_increment': 15663105,
            'priority_stream_id': 1,
            'priority_exclusive': True,
            'priority_depends_on_id': 0,
            'priority_weight': 256
        }

    def test_str(self):
        h2_fp = H2Fingerprint.objects.create(**self.h2_fingerprint_data)
        self.assertEquals(str(h2_fp), f'H2Fingerprint -> oid: {h2_fp.oid}')


class DeviceFingerprintTests(TestCase):
    """
    DeviceFingerprint Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.device_fingerprint_data = {
            'device_category': 'mobile',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                           'Chrome/99.0.4844.74 Safari/537.36'),
            'viewport_height': 768,
            'viewport_width': 1024,
        }

    def test_str(self):
        d_fp = DeviceFingerprint.objects.create(**self.device_fingerprint_data)
        self.assertEquals(
            str(d_fp),
            (f'DeviceFingerprint -> user_agent: {d_fp.user_agent}, device_category: {d_fp.device_category}, '
             f'platform: {d_fp.platform}'))


class TLSFingerprintTests(TestCase):
    """
    TLSFingerprint Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.tls_fingerprint_data = {
            'browser': 'Chrome',
        }

    def test_ciphers(self):
        tls_fp = TLSFingerprint.objects.create(**self.tls_fingerprint_data)
        self.assertTrue(':' in tls_fp.ciphers)

    def test_extensions(self):
        tls_fp = TLSFingerprint.objects.create(**self.tls_fingerprint_data)
        self.assertEquals(type(tls_fp.extensions), int)

        tls_fp.extensions = int(Options.OP_NO_TICKET | Options.OP_NO_RENEGOTIATION | Options.OP_ENABLE_MIDDLEBOX_COMPAT)
        self.assertEquals(tls_fp.extensions, 1074806784)


class ProxyTests(TestCase):
    """
    Proxy Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.proxy_data = {
            'url': 'user123:password456@example.com:4582',
            'country': 'US',
            'city': 'dallas',
        }

    def test_str(self):
        proxy = Proxy.objects.create(**self.proxy_data)
        self.assertEqual(str(proxy), 'Proxy -> url: user123:password456@example.com:4582, mode: GENERAL')

    def test_is_on_cooldown(self):
        proxy = Proxy.objects.create(**self.proxy_data)
        self.assertFalse(proxy.is_on_cooldown)

        proxy.set_last_used()
        self.assertTrue(proxy.is_on_cooldown)

    def test_set_last_used(self):
        proxy = Proxy.objects.create(**self.proxy_data)
        self.assertEquals(proxy.used_count, 0)
        self.assertIsNone(proxy.last_used)

        proxy.set_last_used()
        self.assertEquals(proxy.used_count, 1)
        self.assertIsNotNone(proxy.last_used)

    def test_http_url(self):
        proxy = Proxy.objects.create(**self.proxy_data)
        self.assertTrue(proxy.http_url, 'http://user123:password456@example.com:4582')

    def test_https_url(self):
        proxy = Proxy.objects.create(**self.proxy_data)
        self.assertEquals(proxy.https_url, 'https://user123:password456@example.com:4582')




from django.test import TestCase

from djspoofer import const, exceptions
from djspoofer.models import DeviceFingerprint, H2Fingerprint, Fingerprint, Proxy


class FingerprintManagerTests(TestCase):
    """
    FingerprintManager Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.device_fingerprint_data = {
            'browser': 'Chrome',
            'device_category': 'desktop',
            'platform': 'US',
            'screen_height': 1920,
            'screen_width': 1080,
            'user_agent': ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                           'Chrome/99.0.4844.74 Safari/537.36'),
            'viewport_height': 768,
            'viewport_width': 1024,
        }
        cls.ip_fingerprint_data = {
            'city': 'Dallas',
            'country': 'US',
            'isp': 'Spectrum',
            'ip': '194.60.86.250',
        }

    def test_get_random_desktop_fingerprint(self):
        with self.assertRaises(exceptions.DJSpooferError):
            Fingerprint.objects.random_desktop()

        Fingerprint.objects.create(device_fingerprint=DeviceFingerprint.objects.create(**self.device_fingerprint_data))
        self.assertIsNotNone(Fingerprint.objects.random_desktop())


class H2FingerprintManagerTests(TestCase):
    """
    H2FingerprintManager Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.h2_fingerprint_data = {
            'os': 'Windows',
            'browser': 'Chrome',
            'browser_min_major_version': 95,
            'browser_max_major_version': 100,
            'header_table_size': 65536,
            'enable_push': True,
            'max_concurrent_streams': 1000,
            'initial_window_size': 6291456,
            'max_frame_size': 16384,
            'max_header_list_size': 262144,
            'psuedo_header_order': 'm,a,s,p',
            'window_update_increment': 15663105,
            'priority_stream_id': 1,
            'priority_exclusive': True,
            'priority_depends_on_id': 0,
            'priority_weight': 256
        }

    def test_get_by_browser_info(self):
        with self.assertRaises(exceptions.DJSpooferError):
            Fingerprint.objects.random_desktop()

        H2Fingerprint.objects.create(**self.h2_fingerprint_data)
        h2_fingerprint = H2Fingerprint.objects.get_by_browser_info(
            os='Windows',
            browser='Chrome',
            browser_major_version=78
        )
        self.assertIsNone(h2_fingerprint)
        h2_fingerprint = H2Fingerprint.objects.get_by_browser_info(
            os='Windows',
            browser='Chrome',
            browser_major_version=96
        )
        self.assertIsInstance(h2_fingerprint, H2Fingerprint)


class ProxyManagerTests(TestCase):
    """
    ProxyManager Tests
    """

    @classmethod
    def setUpTestData(cls):
        cls.proxy_data = {
            'url': 'user123:password456@example.com:4582',
            'country': 'US',
            'city': 'dallas',
        }
        cls.proxy_data_2 = {
            'url': 'another123:password456@example.com:4582',
            'country': 'US',
            'city': 'dallas',
        }

    def test_get_rotating_proxy(self):
        Proxy.objects.create(**self.proxy_data)
        with self.assertRaises(exceptions.DJSpooferError):
            Proxy.objects.get_rotating_proxy()

        Proxy.objects.create(mode=const.ProxyModes.ROTATING.value, **self.proxy_data_2)
        self.assertIsNotNone(Proxy.objects.get_rotating_proxy())

    def test_get_sticky_proxy(self):
        Proxy.objects.create(**self.proxy_data)
        with self.assertRaises(exceptions.DJSpooferError):
            Proxy.objects.get_sticky_proxy()

        Proxy.objects.create(mode=const.ProxyModes.STICKY.value, **self.proxy_data_2)
        self.assertIsNotNone(Proxy.objects.get_sticky_proxy())

    def test_get_all_urls(self):
        Proxy.objects.create(**self.proxy_data)
        Proxy.objects.create(mode=const.ProxyModes.STICKY.value, **self.proxy_data_2)

        self.assertListEqual(
            sorted(list(Proxy.objects.get_all_urls())),
            sorted([self.proxy_data['url'], self.proxy_data_2['url']])
        )



from django.conf.locale.en import formats as en_formats
from django.contrib import admin
from import_export.admin import ImportExportModelAdmin

from . import const, resources
from .models import Fingerprint, DeviceFingerprint, TLSFingerprint, H2Fingerprint, Proxy, IntoliFingerprint

en_formats.DATETIME_FORMAT = "M d y H:i"


@admin.register(Fingerprint)
class FingerprintAdmin(ImportExportModelAdmin):
    list_display = ['created', ]
    ordering = ['-created']
    search_fields = ['user_agent']

    show_full_result_count = False


@admin.register(DeviceFingerprint)
class DeviceFingerprintAdmin(ImportExportModelAdmin):
    list_display = ['created', 'user_agent']
    ordering = ['-created']
    search_fields = ['user_agent']

    show_full_result_count = False


@admin.register(TLSFingerprint)
class TLSFingerprintAdmin(ImportExportModelAdmin):
    list_display = ['created', 'browser', 'os', 'extensions', 'ciphers']
    list_filter = ('browser', 'os',)
    ordering = ['-created']
    search_fields = ['extensions']

    show_full_result_count = False


@admin.register(H2Fingerprint)
class H2FingerprintAdmin(ImportExportModelAdmin):
    list_display = ['created', 'browser', 'os']
    list_filter = ('browser', 'os', )
    ordering = ['-created']

    show_full_result_count = False


@admin.register(Proxy)
class ProxyAdmin(ImportExportModelAdmin):
    list_display = ['url', 'mode', 'cooldown', 'last_used', 'used_count', 'country', 'city']
    list_filter = ('mode', 'country', 'city')
    ordering = ['mode', 'url']
    search_fields = ['url']

    show_full_result_count = False
    resource_class = resources.ProxyAdminResource

    def set_rotating(self, request, queryset):
        queryset.update(mode=const.ProxyModes.ROTATING.value)

    set_rotating.short_description = 'Set as Rotating'

    def set_sticky(self, request, queryset):
        queryset.update(mode=const.ProxyModes.STICKY.value)

    set_sticky.short_description = 'Set as Sticky'

    def set_general(self, request, queryset):
        queryset.update(mode=const.ProxyModes.GENERAL.value)

    set_general.short_description = 'Set as General'

    actions = [set_rotating, set_sticky, set_general]


@admin.register(IntoliFingerprint)
class IntoliProfileAdmin(ImportExportModelAdmin):
    list_display = ['created', 'device_category', 'platform', 'user_agent', 'weight']
    list_filter = ('device_category', 'platform',)
    ordering = ['-created']
    search_fields = ['user_agent']

    show_full_result_count = False



import logging

from django.conf import settings
from djstarter import decorators

from .exceptions import IncolumitasError

logger = logging.getLogger(__name__)

BASE_URL = settings.INCOLUMITAS_TLS_API_BASE_URL


@decorators.wrap_exceptions(raise_as=IncolumitasError)
def tls_fingerprint(client):
    url = f'{BASE_URL}/fps'

    params = {
        'detail': '1',
    }

    r = client.get(url, params=params)
    r.raise_for_status()
    return TLSFingerprint(r.json())


class TLSFingerprint:
    def __init__(self, data):
        self.ciphers_length = data['ciphers_length']
        self.ec_point_formats = data['ec_point_formats']
        self.extensions = data['extensions']
        self.signature_algorithms_length = data['signature_algorithms_length']
        self.tls_fp = data['tls_fp']



from djstarter import exceptions


class DJSpooferError(exceptions.AppError):
    """All Exceptions in the Spoofer App inherit from this class"""


class ProxyConnectionFailed(DJSpooferError):
    """For any failed proxy connections"""




from unittest import mock

from django.test import TestCase
from httpcore._models import Origin, Request, URL
from httpcore._sync import HTTPConnection, HTTP2Connection
from httpcore.backends import sync

from djspoofer.models import H2Fingerprint


class ConnectionTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.h2_fingerprint_data = {
            'os': 'Windows',
            'browser': 'Chrome',
            'browser_min_major_version': 95,
            'browser_max_major_version': 100,
            'header_table_size': 65536,
            'enable_push': True,
            'max_concurrent_streams': 1000,
            'initial_window_size': 6291456,
            'max_frame_size': 16384,
            'max_header_list_size': 262144,
            'psuedo_header_order': 'm,a,s,p',
            'window_update_increment': 15663105,
            'priority_stream_id': 1,
            'priority_exclusive': True,
            'priority_depends_on_id': 0,
            'priority_weight': 256
        }

    @mock.patch.object(sync, 'SyncStream')
    @mock.patch.object(HTTPConnection, 'handle_request')
    def test_ok(self, mock_handle_request, mock_sync_stream):
        mock_sync_stream.write.return_value = None

        h2_fingerprint = H2Fingerprint.objects.create(**self.h2_fingerprint_data)
        origin = Origin(
            scheme=b'http',
            host=b'www.example.com',
            port=80,
        )
        url = URL(
            url='http://www.example.com'
        )
        req = Request(
            url=url,
            method='GET',
            headers={
                b'host': 'www.example.com',
                b'h2-fingerprint-id': str(h2_fingerprint.oid)
            })
        http2_connection = HTTP2Connection(
            origin=origin,
            stream=mock_sync_stream(),
            keepalive_expiry=10.0
        )

        http2_connection._receive_response = lambda *args, **kwargs: (200, list())
        mock_handle_request.return_value = http2_connection.handle_request(req)



import logging

from django.conf import settings
from djstarter import decorators

from .exceptions import H2Error

logger = logging.getLogger(__name__)


BASE_URL = settings.H2_FINGERPRINT_API_BASE_URL


@decorators.wrap_exceptions(raise_as=H2Error)
def get_h2_fingerprint(client, *args, **kwargs):
    url = f'{BASE_URL}'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return H2FingerprintResponse(r.json())


class H2FingerprintResponse:
    def __init__(self, data):
        self.fingerprint = data['fingerprint']
        self.settings_frame = data['settings_frame']
        self.window_frame = data['window_frame']
        self.priority_frame = data['priority_frame']
        self.pseudo_headers = data['pseudo_headers']
        self.user_agent = data['user_agent']



from django.core.management.base import BaseCommand

from djspoofer.models import Proxy


class Command(BaseCommand):
    help = 'Add Proxy'

    def add_arguments(self, parser):
        parser.add_argument(
            "--proxy-url",
            required=True,
            type=str,
            help="Set the proxy url",
        )

    def handle(self, *args, **kwargs):
        proxy_url = kwargs['proxy_url']
        Proxy.objects.create_rotating_proxy(url=proxy_url)
        self.stdout.write(self.style.MIGRATE_LABEL('Successfully Created rotating Proxy'))



import json
from importlib.resources import open_text
from unittest import mock

import httpx
from django.test import TestCase
from djspoofer.remote.ja3er import exceptions, ja3er_api
from httpx import Request, Response, codes


class DetailsTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.ja3er.resources', 'details.json') as details_json:
            cls.details_json = json.loads(details_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.details_json
        )

        r_details = ja3er_api.details(
            mock_client,
        )
        self.assertEquals(r_details.ssl_version, '771')
        self.assertEquals(
            r_details.ciphers,
            '4865-4866-4867-49195-49199-49196-49200-52393-52392-49171-49172-156-157-47-53'
        )
        self.assertEquals(r_details.ssl_extensions, '0-23-65281-10-11-35-16-5-13-18-51-45-43-27-21')
        self.assertEquals(r_details.elliptic_curve, '29-23-24')
        self.assertEquals(r_details.elliptic_curve_point_format, '0')

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.details_json
        )

        with self.assertRaises(exceptions.Ja3erError):
            ja3er_api.details(
                mock_client,
            )


class SearchTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.ja3er.resources', 'search.json') as search_json:
            cls.search_json = json.loads(search_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.search_json
        )

        r_search = ja3er_api.search(
            mock_client,
            ja3_hash='b32309a26951912be7dba376398abc3b'
        )

        self.assertEquals(len(r_search.stats), 3)

        stats_2 = r_search.stats[1]
        self.assertEquals(stats_2.user_agent, 'Go-http-client/1.1')

        comment_3 = r_search.comments[2]
        self.assertEquals(comment_3.comment, 'challenge accepted')

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.search_json
        )

        with self.assertRaises(exceptions.Ja3erError):
            ja3er_api.search(
                mock_client,
                ja3_hash='b32309a26951912be7dba376398abc3b'
            )



import random

from django.conf import settings

from . import const


class ProxyBuilder:

    def __init__(self, username, password, netloc, **options):
        self._username = username
        self._password = password
        self._netloc = netloc
        self._options = options

    @property
    def http_url(self):
        return f'http://{self._all_options}:{self._password}@{self._netloc}'

    @property
    def _all_options(self):
        return ';'.join([self._username] + [f'{k}={str(v).replace(" ", "")}' for k, v in self._options.items() if v])


def proxy_weighted_country():
    country_weights = getattr(settings, 'PROXYRACK_COUNTRY_WEIGHTS', const.DEFAULT_PROXYRACK_COUNTRY_WEIGHTS)
    countries = [c[0] for c in country_weights]
    weights = [c[1] for c in country_weights]
    return random.choices(population=countries, weights=weights, k=1)[0]


def proxy_weighted_isp():
    isp_weights = getattr(settings, 'PROXYRACK_ISP_WEIGHTS', const.DEFAULT_PROXYRACK_ISP_WEIGHTS)
    isps = [c[0] for c in isp_weights]
    weights = [c[1] for c in isp_weights]
    return random.choices(population=isps, weights=weights, k=1)[0]




from django.core.management.base import BaseCommand
from djspoofer.remote.intoli import tasks


class Command(BaseCommand):
    help = 'Get Intoli Profiles'

    def add_arguments(self, parser):
        parser.add_argument(
            "--max-profiles",
            required=True,
            type=int,
            help="Max Number of Intoli Profiles",
        )
        parser.add_argument(
            "--desktop-only",
            default=False,
            required=False,
            type=bool,
            help="Only Desktop Intoli Profiles",
        )
        parser.add_argument(
            "--os-list",
            required=False,
            nargs='*',
            help="Only Include Intoli Profiles that belong to OS list",
        )

    def handle(self, *args, **kwargs):
        try:
            tasks.get_profiles(**kwargs)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL('Finished getting profiles from Intoli'))



import json
from importlib.resources import open_text
from unittest import mock

import httpx
from django.test import TestCase
from httpx import Request, Response, codes

from djspoofer.remote.howsmyssl import howsmyssl_api


class GetProfilesTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.howsmyssl.resources', 'ssl_check.json') as ssl_check_json:
            cls.ssl_check_json = json.loads(ssl_check_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.ssl_check_json
        )
        r_ssl_check = howsmyssl_api.ssl_check(
            mock_client,
        )

        self.assertTrue(r_ssl_check.session_ticket_supported)
        self.assertListEqual(r_ssl_check.given_cipher_suites, [
            "TLS_GREASE_IS_THE_WORD_3A",
            "TLS_AES_128_GCM_SHA256",
            "TLS_AES_256_GCM_SHA384",
            "TLS_CHACHA20_POLY1305_SHA256",
            "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256",
            "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
            "TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384",
            "TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384",
            "TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256",
            "TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256",
            "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA",
            "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA",
            "TLS_RSA_WITH_AES_128_GCM_SHA256",
            "TLS_RSA_WITH_AES_256_GCM_SHA384",
            "TLS_RSA_WITH_AES_128_CBC_SHA",
            "TLS_RSA_WITH_AES_256_CBC_SHA"
          ])



from django.core.management.base import BaseCommand

from djspoofer import utils
from djspoofer.models import Fingerprint, DeviceFingerprint
from djspoofer.models import IntoliFingerprint


class Command(BaseCommand):
    help = 'Create Desktop Fingerprints'

    def add_arguments(self, parser):
        parser.add_argument(
            "--num_to_create",
            required=True,
            type=int,
            help="Number of Fingerprints to Create",
        )

    def handle(self, *args, **kwargs):
        try:
            num_to_create = kwargs['num_to_create']
            self.create_fingerprints(num_to_create)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully created fingerprints'))

    @staticmethod
    def create_fingerprints(num_to_create):
        for i_fp in IntoliFingerprint.objects.weighted_n_desktop(count=num_to_create):
            ua_parser = utils.UserAgentParser(i_fp.user_agent)
            Fingerprint.objects.create(
                device_fingerprint=DeviceFingerprint.objects.create(
                    browser=ua_parser.browser,
                    browser_major_version=ua_parser.browser_major_version,
                    device_category=i_fp.device_category,
                    os=ua_parser.os,
                    platform=i_fp.platform,
                    screen_height=i_fp.screen_height,
                    screen_width=i_fp.screen_width,
                    user_agent=i_fp.user_agent,
                    viewport_height=i_fp.viewport_height,
                    viewport_width=i_fp.viewport_width,
                )
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['djspoofer',
 'djspoofer.management',
 'djspoofer.management.commands',
 'djspoofer.migrations',
 'djspoofer.remote',
 'djspoofer.remote.h2fingerprint',
 'djspoofer.remote.howsmyssl',
 'djspoofer.remote.incolumitas',
 'djspoofer.remote.intoli',
 'djspoofer.remote.ja3er',
 'djspoofer.remote.proxyrack',
 'djspoofer.tests',
 'djspoofer.tests.djspoofer',
 'djspoofer.tests.h2_fingerprint',
 'djspoofer.tests.h2_fingerprint.resources',
 'djspoofer.tests.howsmyssl',
 'djspoofer.tests.howsmyssl.resources',
 'djspoofer.tests.incolumitas',
 'djspoofer.tests.incolumitas.resources',
 'djspoofer.tests.intoli',
 'djspoofer.tests.intoli.resources',
 'djspoofer.tests.ja3er',
 'djspoofer.tests.ja3er.resources',
 'djspoofer.tests.proxyrack',
 'djspoofer.tests.proxyrack.resources']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.4,<4.0.0',
 'Faker>=8.12.1,<8.13.0',
 'coverage>=6.2,<7.0',
 'dj-starter>=0.1.11,<0.2.0',
 'django-environ>=0.8.1,<0.9.0',
 'django-import-export>=2.8.0,<3.0.0',
 'httpx[http2]>=0.22.0,<0.23.0',
 'psycopg2-binary>=2.9.2,<3.0.0',
 'python-dotenv>=0.19.0,<0.20.0',
 'ua-parser>=0.10.0,<0.11.0',
 'urllib3>=1.26.9,<2.0.0',
 'wheel>=0.36.2,<0.37.0']

setup_kwargs = {
    'name': 'dj-spoofer',
    'version': '0.1.0',
    'description': 'Django + Web Scraping Made Easy',
    'long_description': None,
    'author': 'Adrian',
    'author_email': 'adrian@rydeas.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/adrianmeraz/dj-spoofer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class DJSpooferConfig(AppConfig):
    name = 'djspoofer'
    verbose_name = 'DJ Spoofer App'

    def ready(self):
        from httpcore._sync.http2 import HTTP2Connection
        from djspoofer import connections

        # Monkey patching to allow for dynamic h2 settings frame
        HTTP2Connection._send_connection_init = connections._send_connection_init
        HTTP2Connection._send_request_headers = connections._send_request_headers
        logger.debug('Monkey patched HTTP2Connection._send_connection_init')



from django.core.management.base import BaseCommand
from djstarter import utils

from djspoofer.clients import DesktopChromeClient
from djspoofer.remote.ja3er import ja3er_api


class Command(BaseCommand):
    help = 'Ja3er Check'

    def handle(self, *args, **kwargs):
        try:
            with DesktopChromeClient() as chrome_client:
                self.show_ja3er_details(chrome_client)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully got ja3er details'))

    def show_ja3er_details(self, client):
        r_json = ja3er_api.details(client)
        self.stdout.write(utils.eye_catcher_line('JA3 Details'))
        self.stdout.write(utils.pretty_dict(r_json))

        r_search = ja3er_api.search(client, ja3_hash=r_json.ja3_hash)
        self.stdout.write(utils.eye_catcher_line('JA3 Hash Search'))
        self.stdout.write(utils.pretty_dict(vars(r_search)))






import logging

from django.conf import settings
from djstarter import decorators

from .exceptions import HowsMySSLError

logger = logging.getLogger(__name__)

BASE_URL = settings.HOWSMYSSL_API_BASE_URL


@decorators.wrap_exceptions(raise_as=HowsMySSLError)
def ssl_check(client, *args, **kwargs):
    url = f'{BASE_URL}/a/check'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return SSLCheckResponse(r.json())


class SSLCheckResponse:

    def __init__(self, data):
        self.given_cipher_suites = data['given_cipher_suites']
        self.ephemeral_keys_supported = data['ephemeral_keys_supported']
        self.session_ticket_supported = data['session_ticket_supported']
        self.tls_compression_supported = data['tls_compression_supported']
        self.unknown_cipher_suite_supported = data['unknown_cipher_suite_supported']
        self.beast_vuln = data['beast_vuln']
        self.able_to_detect_n_minus_one_splitting = data['able_to_detect_n_minus_one_splitting']
        self.insecure_cipher_suites = data['insecure_cipher_suites']
        self.tls_version = data['tls_version']
        self.rating = data['rating']



from django.conf import settings
from django.core.management.base import BaseCommand
from djstarter import utils

from djspoofer.clients import DesktopChromeClient
from djspoofer.remote.incolumitas import incolumitas_tcpip_api
from djspoofer.remote.proxyrack import utils as pr_utils


class Command(BaseCommand):
    help = 'Get TCP/IP Fingerprint'

    def add_arguments(self, parser):
        parser.add_argument(
            "--proxy-url",
            required=True,
            type=str,
            help="Set the proxy url",
        )
        parser.add_argument(
            "--proxy-args",
            required=False,
            nargs='*',
            help="Set the proxy password",
        )

    def handle(self, *args, **kwargs):
        proxy_builder = pr_utils.ProxyBuilder(
            netloc=kwargs.pop('proxy_url'),
            password=settings.PROXY_PASSWORD,
            username=settings.PROXY_USERNAME,
            **self.proxy_options(kwargs.get('proxy_args', list())),
        )
        try:
            with DesktopChromeClient(proxy_url=proxy_builder.http_url) as client:
                r_tcpip = incolumitas_tcpip_api.tcpip_fingerprint(client)
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(utils.pretty_dict(r_tcpip))
            self.stdout.write(self.style.MIGRATE_LABEL('Finished getting TCP/IP Fingerprint'))

    @staticmethod
    def proxy_options(proxy_args):
        return {args.split('=')[0]: args.split('=')[1] for args in proxy_args}



import logging
import random

from faker import Faker
from ua_parser import user_agent_parser

from . import models, providers

logger = logging.getLogger(__name__)


fake = Faker('en_US')
fake.add_provider(providers.UsernameProvider)
fake.add_provider(providers.PhoneNumberProvider)


class FakeProfile:
    MIN_PWD_LEN = 6

    def __init__(self, username=None):
        self.username = username or fake.username()
        self.gender = random.choice(['M', 'F'])
        self.first_name = fake.first_name_male() if self.gender == 'M' else fake.first_name_female()
        self.last_name = fake.last_name()
        self.dob = fake.date_of_birth(minimum_age=18, maximum_age=60)
        self.contact_email = f'{fake.username()}@{fake.free_email_domain()}'
        self.addr_state = fake.state_abbr()
        self.us_phone_number = fake.us_e164()
        self.password = fake.password(length=12, special_chars=True, digits=True, upper_case=True, lower_case=True)

    def __str__(self):
        return f'FakeProfile -> username: {self.username}, full_name: {self.full_name}'

    @property
    def full_gender(self):
        return 'MALE' if self.gender == 'M' else 'FEMALE'

    @property
    def full_name(self):
        return f'{self.first_name} {self.last_name}'

    @property
    def dob_yyyymmdd(self):
        return self.dob.strftime('%Y-%m-%d')


class UserAgentParser:
    class UserAgent:
        def __init__(self, data):
            self.data = data

        @property
        def family(self):
            return self.data['family']

        @property
        def major(self):
            return self.data['major']

        @property
        def minor(self):
            return self.data['minor']

        @property
        def patch(self):
            return self.data['patch']

    class OS(UserAgent):
        pass

    def __init__(self, user_agent):
        self.ua_parser = user_agent_parser.Parse(user_agent)
        self._user_agent = self.UserAgent(self.ua_parser['user_agent'])
        self._os = self.OS(self.ua_parser['os'])

    @property
    def browser(self):
        return self._user_agent.family

    @property
    def browser_major_version(self):
        return self._user_agent.major

    @property
    def os(self):
        return self._os.family

    def __str__(self):
        return f'UserAgentParser -> {self.ua_parser}'


class H2HashParser:
    class SettingsFrame:
        def __init__(self, data):
            self._data = data
            pairs = [pair.split(':') for pair in data.split(';')]
            kv_map = {str(k): int(v) for k, v in pairs}
            self.header_table_size = kv_map.get('1')
            self.push_enabled = kv_map.get('2')
            self.max_concurrent_streams = kv_map.get('3')
            self.initial_window_size = kv_map.get('4')
            self.max_frame_size = kv_map.get('5')
            self.max_header_list_size = kv_map.get('6')

    class PriorityFrame:
        def __init__(self, data):
            parts = [int(v) for v in data.split(':')]
            self.stream_id = parts[0]
            self.is_exclusive = parts[1]
            self.depends_on_id = parts[2]
            self.weight = parts[3]

    def __init__(self, hash):
        parts = hash.split('|')
        self.settings_frame = self.SettingsFrame(parts[0])
        self.window_frame = int(parts[1] or 0)
        self.priority_frame = self.PriorityFrame(parts[2])
        self.pseudo_headers = parts[3]


def h2_hash_to_h2_fingerprint(os, browser, h2_hash, browser_min_major_version=None, browser_max_major_version=None):
    h2_parser = H2HashParser(hash=h2_hash)
    return models.H2Fingerprint.objects.create(
        browser=browser,
        os=os,
        browser_min_major_version=browser_min_major_version,
        browser_max_major_version=browser_max_major_version,
        header_table_size=h2_parser.settings_frame.header_table_size,
        enable_push=bool(h2_parser.settings_frame.push_enabled),
        max_concurrent_streams=h2_parser.settings_frame.max_concurrent_streams,
        initial_window_size=h2_parser.settings_frame.initial_window_size,
        max_frame_size=h2_parser.settings_frame.max_frame_size,
        max_header_list_size=h2_parser.settings_frame.max_header_list_size,
        psuedo_header_order=h2_parser.pseudo_headers,
        window_update_increment=h2_parser.window_frame,
        priority_stream_id=h2_parser.priority_frame.stream_id,
        priority_exclusive=h2_parser.priority_frame.is_exclusive,
        priority_depends_on_id=h2_parser.priority_frame.depends_on_id,
        priority_weight=h2_parser.priority_frame.weight,
    )


def proxy_dict(proxy_url):
    if proxy_url:
        return {
            'http://': proxy_url,
            'https://': proxy_url,
        }
    return dict()



import logging

from djstarter import decorators

from .exceptions import Ja3erError

logger = logging.getLogger(__name__)

BASE_URL = 'https://ja3er.com'


@decorators.wrap_exceptions(raise_as=Ja3erError)
def details(client, *args, **kwargs):
    url = f'{BASE_URL}/json'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return DetailsResponse(r.json())


class DetailsResponse:

    def __init__(self, data):
        self.data = data
        self.ja3_hash = data['ja3_hash']
        self.ja3 = data['ja3']
        self.ja3_parts = self.ja3.split(',')
        self.user_agent = data['User-Agent']

    @property
    def ssl_version(self):
        return self.ja3_parts[0]

    @property
    def ciphers(self):
        return self.ja3_parts[1]

    @property
    def ssl_extensions(self):
        return self.ja3_parts[2]

    @property
    def elliptic_curve(self):
        return self.ja3_parts[3]

    @property
    def elliptic_curve_point_format(self):
        return self.ja3_parts[4]


@decorators.wrap_exceptions(raise_as=Ja3erError)
def search(client, ja3_hash, *args, **kwargs):
    url = f'{BASE_URL}/search/{ja3_hash}'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return SearchResponse(r.json())


class SearchResponse:

    class Stats:
        def __init__(self, data):
            self.user_agent = data['User-Agent']
            self.count = data['Count']
            self.last_seen = data['Last_seen']

    class Comment:
        def __init__(self, data):
            self.comment = data['Comment']
            self.reported = data['Reported']

    def __init__(self, data):
        self.stats = [self.Stats(s) for s in data if s.get('User-Agent')]
        self.comments = [self.Comment(c) for c in data if c.get('Comment')]



import random

from django.db import models
from django.db import transaction
from django.db.models import F, Q
from django.utils import timezone

from djspoofer import exceptions
from djspoofer.remote.intoli import exceptions as intoli_exceptions
from . import const


class GeolocationManager(models.Manager):
    pass


class H2FingerprintManager(models.Manager):
    def get_by_browser_info(self, os, browser, browser_major_version):
        q = Q(
            os=os,
            browser=browser,
            browser_min_major_version__lte=browser_major_version,
            browser_max_major_version__gte=browser_major_version
        )
        return super().get_queryset().filter(q).first()

    def get_by_oid(self, oid):
        return super().get_queryset().get(oid=oid)


class IPManager(models.Manager):
    pass


class TLSFingerprintManager(models.Manager):
    pass


class FingerprintManager(models.Manager):
    def desktop_only(self):
        q = Q(device_fingerprint__device_category='desktop', device_fingerprint__browser__in=const.SUPPORTED_BROWSERS)
        return super().get_queryset().filter(q)

    def random_desktop(self):
        try:
            return self.desktop_only().order_by('?')[0]
        except Exception:
            raise exceptions.DJSpooferError('No Desktop Fingerprints Exist. Did you run the djspoofer_init command?')


class ProxyManager(models.Manager):

    def create_general_proxy(self, *args, **kwargs):
        return self.create(mode=const.ProxyModes.GENERAL.value, *args, **kwargs)

    def create_rotating_proxy(self, *args, **kwargs):
        return self.create(mode=const.ProxyModes.ROTATING.value, *args, **kwargs)

    def create_sticky_proxy(self, *args, **kwargs):
        return self.create(mode=const.ProxyModes.STICKY.value, *args, **kwargs)

    def get_rotating_proxy(self):
        q_filter = Q(mode=const.ProxyModes.ROTATING.value)
        try:
            return super().get_queryset().filter(q_filter)[0]
        except IndexError:
            raise exceptions.DJSpooferError('No rotating proxy is available. Did you run the djspoofer_init command?')

    def get_sticky_proxy(self):
        with transaction.atomic():
            q = Q(mode=const.ProxyModes.STICKY.value)
            q &= (Q(last_used__lt=timezone.now() - F('cooldown')) | Q(last_used=None))

            try:
                sticky_proxy = super().get_queryset().select_for_update(skip_locked=True).order_by(
                    F('last_used').asc(nulls_first=True)).filter(q)[0]
            except IndexError:
                raise exceptions.DJSpooferError('No sticky proxy is available')

            sticky_proxy.set_last_used()
            return sticky_proxy

    def get_all_urls(self):
        return super().get_queryset().values_list('url', flat=True)


class IntoliFingerprintManager(models.Manager):
    def all_oids(self):
        return super().get_queryset().values_list('oid', flat=True)

    def all_user_agents(self):
        return super().get_queryset().values_list('user_agent', flat=True)

    def all_desktop(self):
        return super().get_queryset().filter(
            device_category='desktop',
            browser__in=const.SUPPORTED_BROWSERS,
            os__in=const.SUPPORTED_OS,
        )

    def all_mobile(self):
        return super().get_queryset().filter(device_category='mobile')

    def random_desktop(self):
        try:
            return self.all_desktop().order_by('?')[0]
        except Exception:
            raise intoli_exceptions.IntoliError(
                'No Desktop Intoli Fingerprints Exist. Did you run the "intoli_get_profiles" command?'
            )

    def random_mobile(self):
        try:
            return self.all_mobile().order_by('?')[0]
        except Exception:
            raise intoli_exceptions.IntoliError(
                'No Mobile Intoli Fingerprints Exist. Did you run the "intoli_get_profiles" command?'
            )

    def weighted_n_desktop(self, count=1):
        try:
            desktop_profiles = self.all_desktop()
            weights = [float(p.weight) for p in desktop_profiles]
            return random.choices(population=desktop_profiles, weights=weights, k=count)
        except IndexError:
            raise intoli_exceptions.IntoliError('No Desktop Profiles Exist')

    def weighted_n_mobile(self, count=1):
        try:
            mobile_profiles = self.all_mobile()
            weights = [float(p.weight) for p in mobile_profiles]
            return random.choices(population=mobile_profiles, weights=weights, k=count)
        except IndexError:
            raise intoli_exceptions.IntoliError('No Mobile Profiles Exist')

    def bulk_delete(self, oids):
        return super().get_queryset().filter(oid__in=oids).delete()



# Generated by Django 3.2.13 on 2022-05-06 20:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djspoofer', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='proxy',
            name='url',
            field=models.TextField(),
        ),
    ]



import json
from importlib.resources import open_text
from unittest import mock

import httpx
from django.test import TestCase
from httpx import Request, Response, codes

from djspoofer.remote.incolumitas import exceptions, incolumitas_api


class IPFingerprintTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.request = Request(url='', method='')  # Must add a non null request to avoid raising Runtime exception
        with open_text('djspoofer.tests.incolumitas.resources', 'ip_fingerprint.json') as ip_fingerprint_json:
            cls.ip_fingerprint_json = json.loads(ip_fingerprint_json.read())

    @mock.patch.object(httpx, 'Client')
    def test_ok(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.OK,
            json=self.ip_fingerprint_json
        )

        r_ip_fingerprint = incolumitas_api.ip_fingerprint(
            mock_client,
            ip_addr='76.187.119.132'
        )
        self.assertEquals(r_ip_fingerprint.ip, '76.187.119.132')

    @mock.patch.object(httpx, 'Client')
    def test_400(self, mock_client):
        mock_client.get.return_value = Response(
            request=self.request,
            status_code=codes.BAD_REQUEST,
            json=self.ip_fingerprint_json
        )

        with self.assertRaises(exceptions.IncolumitasError):
            incolumitas_api.ip_fingerprint(
                mock_client,
            )



from django.core.management.base import BaseCommand
from djstarter import utils

from djspoofer.clients import DesktopChromeClient
from djspoofer.remote.howsmyssl import howsmyssl_api


class Command(BaseCommand):
    help = 'SSL Check'

    def handle(self, *args, **kwargs):
        try:
            with DesktopChromeClient() as chrome_client:
                r_check = howsmyssl_api.ssl_check(chrome_client)
                self.stdout.write(utils.pretty_dict(r_check))
        except Exception as e:
            self.stdout.write(self.style.ERROR(f'Error while running command:\n{str(e)}'))
            raise e
        else:
            self.stdout.write(self.style.MIGRATE_LABEL(f'Successfully got ssl check details'))



import logging
import httpx
from django.conf import settings

from djstarter import decorators

from djspoofer.remote.proxyrack import exceptions

logger = logging.getLogger(__name__)

BASE_URL = settings.PROXYRACK_API_BASE_URL


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def active_connections(client, *args, **kwargs):
    url = f'{BASE_URL}/active_conns'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return ActiveConnectionsResponse(r.json())


class ActiveConnectionsResponse:
    class Connection:
        def __init__(self, data):
            self.create_time = data['createTime']
            self.dest_addr = data['destAddr']
            self.source_addr = data['sourceAddr']

    def __init__(self, data):
        self.connections = [self.Connection(c) for c in data]


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def cities(client, *args, **kwargs):
    url = f'{BASE_URL}/cities'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return CitiesResponse(r.json())


class CitiesResponse:
    def __init__(self, data):
        self.cities = data


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def countries(client, *args, **kwargs):
    url = f'{BASE_URL}/countries'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return CountriesResponse(r.json())


class CountriesResponse:
    def __init__(self, data):
        self.countries = data


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def isps(client, country, *args, **kwargs):
    url = f'{BASE_URL}/countries/{country}/isps'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return ISPSResponse(r.json())


class ISPSResponse:
    def __init__(self, data):
        self.isps = data


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def country_ip_count(client, country, *args, **kwargs):
    url = f'{BASE_URL}/countries/{country}/count'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return r.text


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def generate_temp_api_key(client, expiration_seconds, *args, **kwargs):
    url = f'{BASE_URL}/passwords'

    params = {
        'expirationSeconds': expiration_seconds,
    }

    r = client.post(url, params=params, *args, **kwargs)
    r.raise_for_status()
    return GenerateTempApiKeyResponse(r.json())


class GenerateTempApiKeyResponse:
    class Password:
        def __init__(self, data):
            self.expiration_seconds = data['expirationSeconds']
            self.password = data['password']

    def __init__(self, data):
        self.password = self.Password(data['password'])
        self.success = data['success']

    @property
    def api_key(self):
        return self.password.password


status_errors_map = {
    407: exceptions.ProxyNotAuthenticated,
    560: exceptions.GeoLocationNotFound,
    561: exceptions.ProxyUnreachable,
    562: exceptions.ProxyNotFound,
    564: exceptions.ProxyNotOnline
}


def is_valid_proxy(proxies):
    url = 'https://api.ipify.org?format=json'
    try:
        r = httpx.get(url, proxies=proxies)
        return not r.is_error
    except httpx.TransportError as e:
        logger.warning(f'Error while testing proxies. Error: {str(e)}')
        return False


@decorators.wrap_exceptions(raise_as=exceptions.ProxyRackError)
def stats(client, *args, **kwargs):
    """
    Gets statistics of current proxy
    Note: An endpoint must be hit first, or data will be blank
    :param client:
    :param args:
    :param kwargs:
    :return:
    """

    url = f'{BASE_URL}/stats'
    r = client.get(url, *args, **kwargs)
    r.raise_for_status()
    return StatsResponse(r.json())


class StatsResponse:
    class IPInfo:
        class Fingerprint:
            def __init__(self, data):
                self.osName = data['osName']

        def __init__(self, data):
            self.city = data['city']
            self.country = data['country']
            self.fingerprint = self.Fingerprint(data['fingerprint'])
            self.ip = data['ip']
            self.isp = data['isp']
            self.online = data['online']
            self.proxyId = data.get('proxyId')

    def __init__(self, data):
        self.ipinfo = self.IPInfo(data['ipinfo'])
        self.thread_limit = data['threadLimit']
